//-----------------------------------------------------------------------------
// AtgMessages.h
//
// Defines basic data structures using for passing messages. Used with the
// AtgAsyncTaskQueue class
//
// Xbox Advanced Technology Group.
// Copyright (C) Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------
#pragma once
#include "AtgTasks.h"

namespace ATG
{

//
// Message Identifiers
// The high word of a message identifier gives the message area
// The low word of a message identifier gives the message number
//

#define MSGID(Area, Number)         (DWORD)((WORD)(Area) << 16 | (WORD)(Number))
#define MSG_AREA(msgid)             (((msgid) >> 16) & 0xFFFF)
#define MSG_NUMBER(msgid)           ((msgid) & 0xFFFF)

//
// Message Areas
//

#define MSGAREA_NONE                (0x0000)
#define MSGAREA_SESSIONS            (0x0001)
#define MSGAREA_APP                 (0x0002)

//
// Session-related Messages
//
// IMPORTANT: The actual values of the messages are not important,
//            but the relative order is. When adding, deleting, or
//            changing messages, please make sure that the
//            relative order is in the order that these messages
//            would be processed if they were added to a priority
//            queue. The lower the id, the higher the priority.
//

#define MSG_SESSION_CREATE                  MSGID(MSGAREA_SESSIONS, 0x0001)
#define MSG_SESSION_ADDLOCAL                MSGID(MSGAREA_SESSIONS, 0x0002)
#define MSG_SESSION_ADDREMOTE               MSGID(MSGAREA_SESSIONS, 0x0003)
#define MSG_SESSION_REMOVELOCAL             MSGID(MSGAREA_SESSIONS, 0x0004)
#define MSG_SESSION_REMOVEREMOTE            MSGID(MSGAREA_SESSIONS, 0x0005)
#define MSG_SESSION_START                   MSGID(MSGAREA_SESSIONS, 0x0006)
#define MSG_SESSION_MODIFYFLAGS             MSGID(MSGAREA_SESSIONS, 0x0007)
#define MSG_SESSION_MODIFYSKILL             MSGID(MSGAREA_SESSIONS, 0x0008)
#define MSG_SESSION_REGISTERARBITRATION     MSGID(MSGAREA_SESSIONS, 0x0009)
#define MSG_SESSION_WRITESTATS              MSGID(MSGAREA_SESSIONS, 0x000A)
#define MSG_SESSION_END                     MSGID(MSGAREA_SESSIONS, 0x000B)
#define MSG_SESSION_MIGRATE                 MSGID(MSGAREA_SESSIONS, 0x000C)
#define MSG_SESSION_DELETE                  MSGID(MSGAREA_SESSIONS, 0x000D)

//-----------------------------------------------------------------------------
//  Session API message structures
//-----------------------------------------------------------------------------
struct SessionCreateMsg : public MessageBase
{
    DWORD           m_dwFlags;
    DWORD           m_dwMaxPublicSlots;
    DWORD           m_dwMaxPrivateSlots;
    DWORD           m_dwUserIndex;
};

struct SessionStartMsg : public MessageBase
{
};

struct SessionEndMsg : public MessageBase
{
};

struct SessionDeleteMsg : public MessageBase
{
};

struct SessionModifyFlagsMsg : public MessageBase
{
    DWORD           m_dwFlags;
    BOOL            m_bClearFlags;
};

struct ArbitrationRegisterMsg : public MessageBase
{
};

struct SessionMigrateMsg : public MessageBase
{
};

struct AddLocalPlayersMsg : public MessageBase
{
    DWORD       m_dwUserCount;
    DWORD       m_pdwUserIndices[XUSER_MAX_COUNT];
    BOOL        m_pfPrivateSlots[XUSER_MAX_COUNT];

    AddLocalPlayersMsg( const DWORD  dwUserCount,
                        const DWORD* pdwUserIndices,
                        const BOOL*  pfPrivateSlots )
    {
        m_dwUserCount       = dwUserCount;

        memcpy( m_pdwUserIndices, pdwUserIndices, dwUserCount * sizeof(DWORD) );
        memcpy( m_pfPrivateSlots, pfPrivateSlots, dwUserCount * sizeof(BOOL) );
    }
};

struct AddRemotePlayersMsg : public MessageBase
{
    DWORD       m_dwXuidCount;
    XUID        m_pXuids[XUSER_MAX_COUNT];
    BOOL        m_pfPrivateSlots[XUSER_MAX_COUNT];

    AddRemotePlayersMsg( const DWORD dwXuidCount,
                         const XUID* pXuids,
                         const BOOL* pfPrivateSlots )
    {
        m_dwXuidCount       = dwXuidCount;
        
        memcpy( m_pXuids, pXuids, dwXuidCount * sizeof(XUID) );
        memcpy( m_pfPrivateSlots, pfPrivateSlots, dwXuidCount * sizeof(BOOL) );
    }
};

struct RemoveLocalPlayersMsg : public MessageBase
{
    DWORD       m_dwUserCount;
    DWORD       m_pdwUserIndices[XUSER_MAX_COUNT];
    BOOL        m_pfPrivateSlots[XUSER_MAX_COUNT];

    RemoveLocalPlayersMsg( const DWORD  dwUserCount,
                           const DWORD* pdwUserIndices,
                           const BOOL*  pfPrivateSlots )
    {
        m_dwUserCount       = dwUserCount;

        memcpy( m_pdwUserIndices, pdwUserIndices, dwUserCount * sizeof(DWORD) );
        memcpy( m_pfPrivateSlots, pfPrivateSlots, dwUserCount * sizeof(BOOL) );
    }
};

struct RemoveRemotePlayersMsg : public MessageBase
{
    DWORD       m_dwXuidCount;
    XUID        m_pXuids[XUSER_MAX_COUNT];
    BOOL        m_pfPrivateSlots[XUSER_MAX_COUNT];

    RemoveRemotePlayersMsg( const DWORD dwXuidCount,
                            const XUID* pXuids,
                            const BOOL* pfPrivateSlots )
    {
        m_dwXuidCount       = dwXuidCount;

        memcpy( m_pXuids, pXuids, dwXuidCount * sizeof(XUID) );
        memcpy( m_pfPrivateSlots, pfPrivateSlots, dwXuidCount * sizeof(BOOL) );
    }
};

struct ModifySkillMsg : public MessageBase
{
    DWORD       m_dwXuidCount;
    XUID        m_pXuids[X_STATS_MAX_USER_COUNT];

    ModifySkillMsg( const DWORD dwNumXuids, const XUID* pXuids )
    {
        m_dwXuidCount   = dwNumXuids;

        memcpy( (void*)m_pXuids, pXuids, dwNumXuids * sizeof(XUID) );
    }
};

struct WriteStatsMsg : public MessageBase
{
    XUID                            m_xuid;
    DWORD                           m_dwNumViews;
    const XSESSION_VIEW_PROPERTIES* m_pViews;

    WriteStatsMsg( XUID xuid,
                   DWORD dwNumViews,
                   const XSESSION_VIEW_PROPERTIES *pViews)
    {
        // Memory pre-allocated for us, so don't need to allocate here. Also, 
        // don't destroy!
        m_xuid          = xuid;
        m_dwNumViews    = dwNumViews;
        m_pViews        = pViews;        
    }
};

} // namespace ATG
