//-----------------------------------------------------------------------------
// AtgSessionManager.h
//
// Class for managing all session-related activity for a single
// XSession instance
//
// Xbox Advanced Technology Group.
// Copyright (C) Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------
#pragma once
#include "AtgMessageRouter.h"
#include "AtgMessages.h"
#include "AtgTasks.h"
#include "AtgSyncTaskQueue.h"

namespace ATG
{

//--------------------------------------------------------------------------------------
// Callback for MessageRouter
//--------------------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
extern VOID SessionManagerMessageCallback( DWORD dwMessageId, MessageBase* pMsg, VOID* pConsumer );

//--------------------------------------------------------------------------------------
// Constants
//--------------------------------------------------------------------------------------
const DWORD MODIFY_FLAGS_ALLOWED = XSESSION_CREATE_USES_ARBITRATION |
                                   XSESSION_CREATE_INVITES_DISABLED |
                                   XSESSION_CREATE_JOIN_VIA_PRESENCE_DISABLED |
                                   XSESSION_CREATE_JOIN_IN_PROGRESS_DISABLED;

//------------------------------------------------------------------------
// Name: enum SessionState
// Desc: Session state enumeration       
//------------------------------------------------------------------------
enum SessionState
{
    SessionStateNone,
    SessionStateCreate,                 // XSessionCreate task created
    SessionStateCreating,               // XSessionCreate has been called
    SessionStateCreated,                // XSessionCreate has returned
    SessionStateIdle,                   // Session waiting for prompts
    SessionStateWaitingForRegistration, // Session waiting to register with arbitration
    SessionStateRegister,               // XSessionArbitrationRegister task created
    SessionStateRegistering,            // XSessionArbitrationRegister has been called
    SessionStateRegistered,             // XSessionArbitrationRegiste has returned
    SessionStateStart,                  // XSessionStart task created
    SessionStateStarting,               // XSessionStart has been called
    SessionStateInGame,                 // XSessionStart has returned
    SessionStateMigrateHost,            // XSessionMigrateHost task created
    SessionStateMigratingHost,          // XSessionMigrateHost has been called
    SessionStateMigratedHost,           // XSessionMigrateHost has returned
    SessionStateEnd,                    // XSessionEnd task created
    SessionStateEnding,                 // XSessionEnd has been called
    SessionStateEnded,                  // XSessionEnd has returned
    SessionStateDelete,                 // XSessionDelete task created
    SessionStateDeleting,               // XSessionDelete has been called, session is invalid
    SessionStateDeleted,                // XSessionDelete has returned, session is invalid
    SessionStateCount,                  // Count of SessionState enumeration values
};

static const char* g_astrSessionState[SessionStateCount] =
{
    "SessionStateNone",
    "SessionStateCreate",                 // XSessionCreate task created
    "SessionStateCreating",               // XSessionCreate has been called
    "SessionStateCreated",                // XSessionCreate has returned
    "SessionStateIdle",                   // Session waiting for prompts
    "SessionStateWaitingForRegistration", // Session waiting to register with arbitration
    "SessionStateRegister",               // XSessionArbitrationRegister task created
    "SessionStateRegistering",            // XSessionArbitrationRegister has been called
    "SessionStateRegistered",             // XSessionArbitrationRegiste has returned
    "SessionStateStart",                  // XSessionStart task created
    "SessionStateStarting",               // XSessionStart has been called
    "SessionStateInGame",                 // XSessionStart has returned
    "SessionStateMigrateHost",            // XSessionMigrateHost task created
    "SessionStateMigratingHost",          // XSessionMigrateHost has been called
    "SessionStateMigratedHost",           // XSessionMigrateHost has returned
    "SessionStateEnd",                    // XSessionEnd task created
    "SessionStateEnding",                 // XSessionEnd has been called
    "SessionStateEnded",                  // XSessionEnd has returned
    "SessionStateDelete",                 // XSessionDelete task created
    "SessionStateDeleting",               // XSessionDelete has been called, session is invalid
    "SessionStateDeleted"                 // XSessionDelete has returned, session is invalid
};


//------------------------------------------------------------------------
// Name: enum SessionState
// Desc: Slot types for the session       
//------------------------------------------------------------------------
typedef enum _SLOTS
{
    SLOTS_TOTALPUBLIC,
    SLOTS_TOTALPRIVATE,
    SLOTS_FILLEDPUBLIC,
    SLOTS_FILLEDPRIVATE,
    SLOTS_MAX
} SLOTS;

//------------------------------------------------------------------------
// Name: enum SessionCreationType
// Desc: Enum to track the reason for creation of this SessionManager instance
//------------------------------------------------------------------------
enum SessionCreationReason
{
    SessionCreationReasonNone,
    SessionCreationReasonAmbiguous,
    SessionCreationReasonHosting,
    SessionCreationReasonJoinFromSearch
};

//------------------------------------------------------------------------
// Name: struct SessionManagerInitParams
// Desc: Initialization parameters       
//------------------------------------------------------------------------
template <class TaskQueueType>
struct SessionManagerInitParams
{
    SessionCreationReason           m_SessionCreationReason;                
    DWORD                           m_dwSessionFlags;
    BOOL                            m_bIsHost;
    UINT                            m_dwMaxPublicSlots;
    UINT                            m_dwMaxPrivateSlots;
    MessageRouter<TaskQueueType>*   m_pMessageRouter;
};

//--------------------------------------------------------------------------------------
// Name: class SessionManager
// Desc: Class for managing all XSession-related calls for a single LIVE session. 
//--------------------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
class SessionManager
{
#define SESSIONMANAGER_SIGNATURE 0xF00DF00D // Used to validate whether an
                                            // instance pointer is pointing
                                            // to a valid instance
friend VOID SessionManagerMessageCallback
    <TaskQueueType, MessageRouterType>( DWORD dwMessageId,
                                        MessageBase* pMsg,
                                        VOID* pConsumer );

public:

    SessionManager( VOID );
    virtual ~SessionManager( VOID );

    BOOL Initialize( ATG::SessionManagerInitParams<TaskQueueType> initParams );
    BOOL IsInitialized( VOID );

    SessionCreationReason GetSessionCreationReason( VOID ) const;

    DWORD GetSessionOwner( VOID ) const;
    XUID GetSessionOwnerXuid( VOID ) const;
    VOID SetSessionOwner( DWORD dwOwnerController );
    BOOL IsSessionOwner( DWORD dwController ) const;
    BOOL IsSessionOwner( XUID xuidOwner ) const;
    BOOL IsPlayerInSession( XUID xuidPlayer ) const;

    BOOL IsSessionHost( VOID ) const;
    VOID MakeSessionHost( VOID );

    ULONGLONG GetSessionNonce( VOID ) const;
    VOID SetSessionNonce( ULONGLONG qwNonce );

    VOID SetHostInAddr( const IN_ADDR& inaddr );
    IN_ADDR GetHostInAddr( VOID ) const;

    XNKID GetSessionID( VOID ) const;
    __int64 GetSessionIDAsInt( VOID ) const;
    XNKID GetMigratedSessionID( VOID ) const;

    BOOL HasSessionFlags( DWORD dwFlags ) const;
    DWORD GetSessionFlags( VOID ) const;

    VOID GetMaxSlotCounts( DWORD& dwMaxPublicSlots, DWORD& dwMaxPrivateSlots ) const;
    VOID GetFilledSlotCounts( DWORD& dwFilledPublicSlots, DWORD& dwFilledPrivateSlots ) const;

    VOID SetMaxSlotCounts( DWORD dwMaxPublicSlots, DWORD dwMaxPrivateSlots );
    VOID SetFilledSlotCounts( DWORD dwFilledPublicSlots, DWORD dwFilledPrivateSlots );

    const XSESSION_INFO& GetSessionInfo( VOID ) const;
    VOID SetSessionInfo( const XSESSION_INFO& session_info );
    VOID SetNewSessionInfo( const XSESSION_INFO& session_info, const BOOL bIsNewSessionHost );

    VOID SetSessionFlags( DWORD dwFlags, BOOL fClearExisting = FALSE );
    VOID FlipSessionFlags( DWORD dwFlags );
    VOID ClearSessionFlags( DWORD dwFlags );
    DWORD ModifySessionFlags( DWORD dwFlags, BOOL bClearFlags = FALSE );

    WCHAR* GetSessionError( VOID ) const;
    VOID SetSessionError( WCHAR* error );

    SessionState GetSessionState( VOID ) const;
    const char* GetSessionStateString( VOID ) const;
    VOID SwitchToState( SessionState newState );
    VOID SwitchToPreHostMigrationState();

    const PXSESSION_REGISTRATION_RESULTS GetRegistrationResults() const;

    HRESULT ModifySkill( const DWORD dwNumXuids, const XUID* pXuids );

    HRESULT AddLocalPlayers( const DWORD dwUserCount,
                                 const DWORD* pdwUserIndices,
                                 const BOOL* pfPrivateSlots );
    HRESULT RemoveLocalPlayers( const DWORD dwUserCount,
                                      const DWORD* pdwUserIndices,
                                      const BOOL* pfPrivateSlots );
    HRESULT AddRemotePlayers( const DWORD dwXuidCount,
                                    const XUID* pXuids,
                                    const BOOL* pfPrivateSlots );
    HRESULT RemoveRemotePlayers( const DWORD dwXuidCount,
                                         const XUID* pXuids,
                                         const BOOL* pfPrivateSlots );
    HRESULT WriteStats( XUID xuid,
                          DWORD dwNumViews,
                          const XSESSION_VIEW_PROPERTIES *pViews );

    VOID StartQoSListener( BYTE* data, UINT dataLen, DWORD bitsPerSec );

    BOOL IsProcessingOverlappedOperation( VOID ) const;
    DWORD GetOverlappedExtendedError( VOID ) const;

    DWORD GetSignature( VOID ) const { return m_dwSig; } 
    BOOL ValidityCheck( VOID ) const;
    BOOL IsSessionDeleted() const;

    VOID DebugDumpSessionDetails( VOID ) const;

    BOOL DebugValidateExpectedSlotCounts( const BOOL bBreakOnDifferent = TRUE ) const;

protected:
    VOID Reset( VOID );

    HRESULT CreateSession( VOID );
    HRESULT RegisterArbitration( VOID );
    HRESULT StartSession( VOID );
    HRESULT EndSession( VOID );
    HRESULT MigrateSession( VOID );
    HRESULT DeleteSession( VOID );

    DWORD ProcessCreateSessionMessage( MessageBase* pMessage );
    DWORD ProcessStartSessionMessage( MessageBase* pMessage );
    DWORD ProcessModifySessionFlagsMessage( MessageBase* pMessage );
    DWORD ProcessModifySkillMessage( MessageBase* pMessage );
    DWORD ProcessWriteStatsMessage( MessageBase* pMessage );
    DWORD ProcessEndSessionMessage( MessageBase* pMessage );
    DWORD ProcessMigrateHostMessage( MessageBase* pMessage );
    DWORD ProcessDeleteSessionMessage( MessageBase* pMessage );
    DWORD ProcessArbitrationRegisterMessage( MessageBase* pMessage );
    DWORD ProcessAddLocalPlayerMessage( MessageBase* pMessage );
    DWORD ProcessAddRemotePlayerMessage( MessageBase* pMessage );
    DWORD ProcessRemoveLocalPlayerMessage( MessageBase* pMessage );
    DWORD ProcessRemoveRemotePlayerMessage( MessageBase* pMessage );

    VOID SetSessionState( SessionState newState );
    BOOL CheckSessionState( SessionState expectedState ) const;

    VOID StopQoSListener();
    
private:
    SessionManager& operator=( const SessionManager& ref ) {}
    VOID InitializeOverlapped();

    inline
    VOID PushTask( DWORD dwMessage, MessageBase* pMessage );

    inline
    VOID RunTaskNow( DWORD dwMessage, MessageBase* pMessage );

protected:
    const DWORD             m_dwSig;                    // Used for checking that a memory location
                                                        // holds a valid SessionManager instance
    SessionCreationReason   m_SessionCreationReason;    // Creation reason
    BOOL                    m_bIsInitialized;           // Initialized?
    BOOL                    m_bIsHost;                  // Is hosting
    BOOL                    m_bUsingQoS;                // Is the QoS listener enabled
    DWORD                   m_dwOwnerController;        // Which controller created the session
    XUID                    m_xuidOwner;                // XUID of gamer who created the session
    WCHAR*                  m_strSessionError;          // Error message for current session
    MessageRouterType*      m_pMessageRouter;           // MessageRouter for current session
    XOVERLAPPED             m_XOverlapped;              // Overlapped structure for the session
    HANDLE                  m_hTaskCompletedEvent;      // Signallable event for overlapped calls
    ULONGLONG               m_qwSessionNonce;           // Session nonce
    XSESSION_INFO           m_SessionInfo;              // Session ID, key, and host address 
    XSESSION_INFO           m_NewSessionInfo;           // New session ID, key, and host address 
    BOOL                    m_bIsMigratedSessionHost;   // Is hosting migrated session?
    XNKID                   m_migratedSessionID;        // Session ID of migrated session
    HANDLE                  m_hSession;                 // Session handle
    SessionState            m_SessionState;             // Current session state
    SessionState            m_migratedSessionState;     // Session state of migrated session
    DWORD                   m_dwSessionFlags;           // Flags for the current session  exit?
    UINT                    m_Slots[ SLOTS_MAX ];       // Filled/open slots for the session
    IN_ADDR                 m_HostInAddr;               // Host IP address
    
    PXSESSION_REGISTRATION_RESULTS m_pRegistrationResults;
};

}
