#include "AtgMessageRouterImpl.h"
#include "AtgSessionManager.h"

namespace ATG
{

//-----------------------------------------------------------------------------
// Name: SessionManagerMessageCallback
// Desc: Callback for MessageRouter-routed messages
//-----------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
VOID SessionManagerMessageCallback( DWORD dwMessageId,
                                    MessageBase* pMsg,
                                    VOID* pConsumer )
{
    SessionManager<TaskQueueType, MessageRouterType>* pSessionMgr = 
        ( SessionManager<TaskQueueType, MessageRouterType>* )pConsumer;
    
    if( pSessionMgr->m_dwSig != SESSIONMANAGER_SIGNATURE )
    {
        FatalError( "pSessionManager->GetSignature() != SESSIONMANAGER_SIGNATURE !!\n" );
    }

    // Assert if message area is wrong
    assert( MSG_AREA( dwMessageId ) == MSGAREA_SESSIONS );

    switch( dwMessageId )
    {
    case MSG_SESSION_CREATE:
        pSessionMgr->ProcessCreateSessionMessage( pMsg );
        break;
    case MSG_SESSION_ADDLOCAL:
        pSessionMgr->ProcessAddLocalPlayerMessage( pMsg );
        break;
    case MSG_SESSION_ADDREMOTE:
        pSessionMgr->ProcessAddRemotePlayerMessage( pMsg );
        break;
    case MSG_SESSION_REMOVELOCAL:
        pSessionMgr->ProcessRemoveLocalPlayerMessage( pMsg );
        break;
    case MSG_SESSION_REMOVEREMOTE:
        pSessionMgr->ProcessRemoveRemotePlayerMessage( pMsg );
        break;
    case MSG_SESSION_START:
        pSessionMgr->ProcessStartSessionMessage( pMsg );
        break;
    case MSG_SESSION_MODIFYFLAGS:
        pSessionMgr->ProcessModifySessionFlagsMessage( pMsg );
        break;
    case MSG_SESSION_MODIFYSKILL:
        pSessionMgr->ProcessModifySkillMessage( pMsg );
        break;
    case MSG_SESSION_REGISTERARBITRATION:
        pSessionMgr->ProcessArbitrationRegisterMessage( pMsg );
        break;
    case MSG_SESSION_MIGRATE:
        pSessionMgr->ProcessMigrateHostMessage( pMsg );
        break;
    case MSG_SESSION_WRITESTATS:
        pSessionMgr->ProcessWriteStatsMessage( pMsg );
        break;
    case MSG_SESSION_END:
        pSessionMgr->ProcessEndSessionMessage( pMsg );
        break;
    case MSG_SESSION_DELETE:
        pSessionMgr->ProcessDeleteSessionMessage( pMsg );
        break;
    }

    return;
}

//------------------------------------------------------------------------
// Name: SessionManager()
// Desc: Public constructor
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
SessionManager<TaskQueueType, MessageRouterType>::SessionManager() : 
    m_dwSig( SESSIONMANAGER_SIGNATURE )
{
    Reset();
    m_pMessageRouter = NULL;

    // Create a signallable event that we'll use to 
    // signal completion of async XSession* calls
    // and to enforce that the calls are serialized
    const BOOL bManualReset = TRUE; //resettable
    const BOOL bInitialState = TRUE; //signalled

    m_hTaskCompletedEvent = CreateEvent(
                                          NULL,             // SD
                                          bManualReset,     // reset type
                                          bInitialState,    // initial state
                                          NULL              // object name
    );

    if( !m_hTaskCompletedEvent )
    {
        FatalError( "CreateEvent failed! GetLastError: %08x\n", GetLastError() );                    
    }
}

//------------------------------------------------------------------------
// Name: Reset()
// Desc: Resets this SessionManager instance to its creation state
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
VOID SessionManager<TaskQueueType, MessageRouterType>::Reset()
{
    m_bIsInitialized            = FALSE;
    m_SessionCreationReason     = SessionCreationReasonNone;
    m_dwSessionFlags            = 0;
    m_dwOwnerController         = XUSER_MAX_COUNT + 1;
    m_xuidOwner                 = INVALID_XUID;
    m_SessionState              = SessionStateNone;
    m_migratedSessionState      = SessionStateNone;
    m_bIsMigratedSessionHost    = FALSE;
    m_hSession                  = INVALID_HANDLE_VALUE;
    m_qwSessionNonce            = 0;
    m_strSessionError           = NULL;
    m_bUsingQoS                 = FALSE;
    ZeroMemory( &m_SessionInfo, sizeof( XSESSION_INFO ) );
    ZeroMemory( &m_NewSessionInfo, sizeof( XSESSION_INFO ) );
    ZeroMemory( &m_migratedSessionID, sizeof( XNKID ) );    
    ZeroMemory( &m_XOverlapped, sizeof( XOVERLAPPED ) );
}


//------------------------------------------------------------------------
// Name: ~SessionManager()
// Desc: Public destructor
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
SessionManager<TaskQueueType, MessageRouterType>::~SessionManager()
{
    // Close our event for managing overlapped operations
    if( m_hTaskCompletedEvent != INVALID_HANDLE_VALUE && 
        m_hTaskCompletedEvent != NULL )
    {
        XCloseHandle( m_hTaskCompletedEvent );
    }

    // Unregister MessageRouter callback
    if( m_pMessageRouter )
    {
        m_pMessageRouter->UnregisterCallback( this );
    }

    // Close our XSession handle to clean up resources held by it
    if( m_hSession != INVALID_HANDLE_VALUE && 
        m_hSession != NULL )
    {
        XCloseHandle( m_hSession );
    }
}


//------------------------------------------------------------------------
// Name: Initialize()
// Desc: Initializes this SessionManager instance
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
BOOL SessionManager<TaskQueueType, MessageRouterType>::Initialize
    ( SessionManagerInitParams<TaskQueueType> initParams )
{
    if( m_bIsInitialized )
    {
        DebugSpew( "SessionManager already initialized. Reinitializing..\n" );
    }

    m_SessionCreationReason         = initParams.m_SessionCreationReason;
    m_dwSessionFlags                = initParams.m_dwSessionFlags;
    m_bIsHost                       = initParams.m_bIsHost;
    m_Slots[SLOTS_TOTALPUBLIC]      = initParams.m_dwMaxPublicSlots;
    m_Slots[SLOTS_TOTALPRIVATE]     = initParams.m_dwMaxPrivateSlots;
    m_Slots[SLOTS_FILLEDPUBLIC]     = 0;
    m_Slots[SLOTS_FILLEDPRIVATE]    = 0;
    m_pMessageRouter                = initParams.m_pMessageRouter;

    assert( m_pMessageRouter != NULL );
    m_pMessageRouter->RegisterCallback( MSGAREA_SESSIONS, 
                                        SessionManagerMessageCallback<TaskQueueType,MessageRouterType>,
                                        this );

    m_bIsInitialized = TRUE;
    return TRUE;
}

//------------------------------------------------------------------------
// Name: IsInitialized()
// Desc: Queries SessionManager initialization state
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
BOOL SessionManager<TaskQueueType, MessageRouterType>::IsInitialized()
{
    return m_bIsInitialized;
}

//------------------------------------------------------------------------
// Name: InitializeOverlapped
// Desc: Initializes the m_XOverlapped we'll be using for XSession* calls
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
VOID SessionManager<TaskQueueType, MessageRouterType>::InitializeOverlapped()
{
    // Close existing m_hTaskCompletedEvent
    if( m_hTaskCompletedEvent != INVALID_HANDLE_VALUE && 
        m_hTaskCompletedEvent != NULL )
    {
        XCloseHandle( m_hTaskCompletedEvent );
    }
    m_hTaskCompletedEvent = NULL;

    // Create a signallable event that we'll use to 
    // signal completion of async XSession* calls
    // and to enforce that the calls are serialized
    const bool bManualReset = true; //resettable
    const bool bInitialState = false; //non-signalled

    m_hTaskCompletedEvent = CreateEvent(
      NULL, // SD
      bManualReset, // reset type
      bInitialState, // initial state
      NULL// object name
    );

    if( !m_hTaskCompletedEvent )
    {
        FatalError( "CreateEvent failed! GetLastError: %08x\n", GetLastError() );                    
    }

    // Set up our overlapped structure with this event
    ZeroMemory( &m_XOverlapped, sizeof( XOVERLAPPED ) );
    m_XOverlapped.hEvent = m_hTaskCompletedEvent;
}

//------------------------------------------------------------------------
// Name: ValidityCheck()
// Desc: Returns TRUE if this is a valid instance; otherwise false
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
BOOL SessionManager<TaskQueueType, MessageRouterType>::ValidityCheck( VOID ) const
{
    return ( m_dwSig == SESSIONMANAGER_SIGNATURE ) ? TRUE : FALSE;
}

//------------------------------------------------------------------------
// Name: IsSessionDeleted()
// Desc: Returns TRUE if the XSession for this instance
//       has a 0 sessionID; otherwise false
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
BOOL SessionManager<TaskQueueType, MessageRouterType>::IsSessionDeleted( VOID ) const
{
    static const XNKID zero = {0};

    if( m_SessionState == SessionStateDeleted )
    {
        return TRUE;
    }

    BOOL bIsSessionDeleted = FALSE;

    if( m_hSession == INVALID_HANDLE_VALUE || m_hSession == NULL )
    {
        // If the session hasn't been created yet, return false
        if ( m_SessionState < SessionStateCreated )
        {
            bIsSessionDeleted = FALSE;
        }
        else
        {
            bIsSessionDeleted = TRUE;
        }

        return bIsSessionDeleted;
    }

    // Call XSessionGetDetails first time to get back the size 
    // of the results buffer we need
    DWORD cbResults = 0;

    DWORD ret = XSessionGetDetails( m_hSession,
                                    &cbResults,
                                    NULL,
                                    NULL ); 

    if( ( ret != ERROR_INSUFFICIENT_BUFFER ) || ( cbResults == 0 ) )
    {
        FatalError( "Failed on first call to XSessionGetDetails, hr=0x%08x\n", ret );
    }
    
    XSESSION_LOCAL_DETAILS* pSessionDetails = (XSESSION_LOCAL_DETAILS*)new BYTE[ cbResults ];
    if( pSessionDetails == NULL )
    {
        FatalError( "Failed to allocate buffer.\n" );
    }

    // Call second time to fill our results buffer
    ret = XSessionGetDetails( m_hSession,
                              &cbResults,
                              pSessionDetails,
                              NULL ); 

    if( ret != ERROR_SUCCESS )
    {
        FatalError( "XSessionGetDetails failed with error %d\n", ret );
    }

    // Iterate through the returned results
    XNKID sessionID = pSessionDetails->sessionInfo.sessionID;

    if ( !memcmp( &sessionID, &zero, sizeof( sessionID ) ) )
    {
        bIsSessionDeleted = TRUE;
    }

    if ( pSessionDetails )
    {
        delete [] pSessionDetails;
    }

    return bIsSessionDeleted;
}

//--------------------------------------------------------------------------------------
// Inline functions for simplying task creation
//--------------------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
VOID SessionManager<TaskQueueType, MessageRouterType>::PushTask( DWORD dwMessage, MessageBase* pMessage )
{
    ATG::TaskData td;
    td.dwMessage        = dwMessage;
    td.pMessage         = pMessage;
    td.pMsgDestination  = this;
    m_pMessageRouter->Send( td );
}

template <class TaskQueueType, class MessageRouterType>
VOID SessionManager<TaskQueueType, MessageRouterType>::RunTaskNow( DWORD dwMessage, MessageBase* pMessage )
{
    ATG::TaskData td;
    td.dwMessage        = dwMessage;
    td.pMessage         = pMessage;
    td.pMsgDestination  = this;
    m_pMessageRouter->SendImmediate( td );
}

//------------------------------------------------------------------------
// Name: IsProcessingOverlappedOperation()
// Desc: Returns true if this instance is in the middle of 
//       processing an overlapped operation; otherwise false
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
BOOL SessionManager<TaskQueueType, MessageRouterType>::IsProcessingOverlappedOperation() const
{
    return !XHasOverlappedIoCompleted( &m_XOverlapped );
}

//------------------------------------------------------------------------
// Name: GetOverlappedExtendedError()
// Desc: Returns the extended error info from the last completed
//       overlapped operation
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
DWORD SessionManager<TaskQueueType, MessageRouterType>::GetOverlappedExtendedError() const
{
    return XGetOverlappedExtendedError( (XOVERLAPPED *)&m_XOverlapped );
}

//------------------------------------------------------------------------
// Name: IsPlayerInSession()
// Desc: Uses XSessionGetDetails API to determine if a player is in the session
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
BOOL SessionManager<TaskQueueType, MessageRouterType>::IsPlayerInSession( XUID xuidPlayer ) const
{
    if( m_hSession == INVALID_HANDLE_VALUE || m_hSession == NULL )
    {
        DebugSpew( "Session handle invalid. Perhaps session has been deleted?\n" );
        return FALSE;
    }

    // Call XSessionGetDetails first time to get back the size 
    // of the results buffer we need
    DWORD cbResults = 0;

    DWORD ret = XSessionGetDetails( m_hSession,
                                    &cbResults,
                                    NULL,
                                    NULL ); 

    if( ( ret != ERROR_INSUFFICIENT_BUFFER ) || ( cbResults == 0 ) )
    {
        FatalError( "Failed on first call to XSessionGetDetails, hr=0x%08x\n", ret );
    }

    // Increase our buffer size to include an estimate of the maximum number
    // of players allowed in the session
    const DWORD dwMaxMembers = m_Slots[SLOTS_TOTALPUBLIC] + m_Slots[SLOTS_TOTALPRIVATE];
    cbResults += dwMaxMembers * sizeof(XSESSION_MEMBER);
    
    XSESSION_LOCAL_DETAILS* pSessionDetails = (XSESSION_LOCAL_DETAILS*)new BYTE[ cbResults ];
    if( pSessionDetails == NULL )
    {
        FatalError( "Failed to allocate buffer.\n" );
    }

    // Call second time to fill our results buffer
    ret = XSessionGetDetails( m_hSession,
                              &cbResults,
                              pSessionDetails,
                              NULL ); 

    if( ret != ERROR_SUCCESS )
    {
        FatalError( "XSessionGetDetails failed with error %d\n", ret );
    }

    BOOL bIsInSession = FALSE;
    for ( DWORD i = 0; i < pSessionDetails->dwReturnedMemberCount; i++ )
    {
        if( xuidPlayer == pSessionDetails->pSessionMembers[i].xuidOnline )
        {
            bIsInSession = TRUE;
            break;
        }
    }

    if ( pSessionDetails )
    {
        delete [] pSessionDetails;
    }

    return bIsInSession;
}

//------------------------------------------------------------------------
// Name: DebugDumpSessionDetails()
// Desc: Dumps session deletes
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
VOID SessionManager<TaskQueueType, MessageRouterType>::DebugDumpSessionDetails() const
{
    #ifdef _DEBUG

    if( m_hSession == INVALID_HANDLE_VALUE || m_hSession == NULL )
    {
        DebugSpew( "Session handle invalid. Perhaps session has been deleted?\n" );
        return;
    }

    // Call XSessionGetDetails first time to get back the size 
    // of the results buffer we need
    DWORD cbResults = 0;

    DWORD ret = XSessionGetDetails( m_hSession,
                                    &cbResults,
                                    NULL,
                                    NULL ); 

    if( ( ret != ERROR_INSUFFICIENT_BUFFER ) || ( cbResults == 0 ) )
    {
        FatalError( "Failed on first call to XSessionGetDetails, hr=0x%08x\n", ret );
    }

    // Increase our buffer size to include an estimate of the maximum number
    // of players allowed in the session
    const DWORD dwMaxMembers = m_Slots[SLOTS_TOTALPUBLIC] + m_Slots[SLOTS_TOTALPRIVATE];
    cbResults += dwMaxMembers * sizeof(XSESSION_MEMBER);
    
    XSESSION_LOCAL_DETAILS* pSessionDetails = (XSESSION_LOCAL_DETAILS*)new BYTE[ cbResults ];
    if( pSessionDetails == NULL )
    {
        FatalError( "Failed to allocate buffer.\n" );
    }

    // Call second time to fill our results buffer
    ret = XSessionGetDetails( m_hSession,
                              &cbResults,
                              pSessionDetails,
                              NULL ); 

    if( ret != ERROR_SUCCESS )
    {
        FatalError( "XSessionGetDetails failed with error %d\n", ret );
    }

    // Iterate through the returned results
    const XNKID sessionID = pSessionDetails->sessionInfo.sessionID;
    DebugSpew( "***************** DebugDumpSessionDetails *****************\n" \
                    "instance: %p\n" \
                    "sessionID: %016I64X\n" \
                    "Matchmaking session?: %s\n" \
                    "m_hSession: %p\n" \
                    "bIsHost: %d\n" \
                    "sessionstate: %s\n" \
                    "qwSessionNonce: %I64u\n" \
                    "dwUserIndexHost: %d\n" \
                    "dwGameType: %d\n" \
                    "dwGameMode: %d\n" \
                    "dwFlags: %d\n" \
                    "dwMaxPublicSlots: %d\n" \
                    "dwMaxPrivateSlots: %d\n" \
                    "dwAvailablePublicSlots: %d\n" \
                    "dwAvailablePrivateSlots: %d\n" \
                    "dwActualMemberCount: %d\n" \
                    "dwReturnedMemberCount: %d\n" \
                    "xnkidArbitration: %016I64X\n",
                    this,
                    XNKIDToInt64( sessionID ),
                    ( HasSessionFlags( XSESSION_CREATE_USES_MATCHMAKING ) ) ? "Yes" : "No",
                    m_hSession,
                    m_bIsHost,
                    g_astrSessionState[m_SessionState],
                    m_qwSessionNonce,
                    pSessionDetails->dwUserIndexHost,
                    pSessionDetails->dwGameType,
                    pSessionDetails->dwGameMode,
                    pSessionDetails->dwFlags,
                    pSessionDetails->dwMaxPublicSlots,
                    pSessionDetails->dwMaxPrivateSlots,
                    pSessionDetails->dwAvailablePublicSlots,
                    pSessionDetails->dwAvailablePrivateSlots,
                    pSessionDetails->dwActualMemberCount,
                    pSessionDetails->dwReturnedMemberCount,
                    XNKIDToInt64( pSessionDetails->xnkidArbitration ) );

    for ( DWORD i = 0; i < pSessionDetails->dwReturnedMemberCount; i++ )
    {
        DebugSpew( "***************** XSESSION_MEMBER %d *****************\n" \
                   "Online XUID: %I64u\n" \
                   "dwUserIndex: %d\n" \
                   "dwFlags: %d\n",
                   i,
                   pSessionDetails->pSessionMembers[i].xuidOnline,
                   pSessionDetails->pSessionMembers[i].dwUserIndex,
                   pSessionDetails->pSessionMembers[i].dwFlags );    
    }

    if ( pSessionDetails )
    {
        delete [] pSessionDetails;
    }

    #endif
}

//------------------------------------------------------------------------
// Name: DebugValidateExpectedSlotCounts()
// Desc: Makes sure slot counts are consistent with what is stored with
//       the XSession
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
BOOL SessionManager<TaskQueueType, MessageRouterType>::DebugValidateExpectedSlotCounts( const BOOL bBreakOnDifferent ) const
{
    BOOL bSlotCountsDiffer = FALSE;

    #ifdef _DEBUG

    if( m_hSession == INVALID_HANDLE_VALUE || m_hSession == NULL )
    {
        DebugSpew( "Session handle invalid. Perhaps session has been deleted?\n" );
        return !bSlotCountsDiffer;
    }

    // Call XSessionGetDetails first time to get back the size 
    // of the results buffer we need
    DWORD cbResults = 0;

    DWORD ret = XSessionGetDetails( m_hSession,
                                    &cbResults,
                                    NULL,
                                    NULL ); 

    if( ( ret != ERROR_INSUFFICIENT_BUFFER ) || ( cbResults == 0 ) )
    {
        FatalError( "Failed on first call to XSessionGetDetails, hr=0x%08x\n", ret );
    }
    
    XSESSION_LOCAL_DETAILS* pSessionDetails = (XSESSION_LOCAL_DETAILS*)new BYTE[ cbResults ];
    if( pSessionDetails == NULL )
    {
        FatalError( "Failed to allocate buffer.\n" );
    }

    // Call second time to fill our results buffer
    ret = XSessionGetDetails( m_hSession,
                              &cbResults,
                              pSessionDetails,
                              NULL ); 

    if( ret != ERROR_SUCCESS )
    {
        FatalError( "XSessionGetDetails failed with error %d\n", ret );
    }

    const XNKID sessionID = pSessionDetails->sessionInfo.sessionID;

    const DWORD dwMaxPublicSlots        = m_Slots[SLOTS_TOTALPUBLIC];
    const DWORD dwMaxPrivateSlots       = m_Slots[SLOTS_TOTALPRIVATE];
    const DWORD dwAvailablePublicSlots  = m_Slots[SLOTS_TOTALPUBLIC] - m_Slots[SLOTS_FILLEDPUBLIC];
    const DWORD dwAvailablePrivateSlots = m_Slots[SLOTS_TOTALPRIVATE] - m_Slots[SLOTS_FILLEDPRIVATE];

    bSlotCountsDiffer = ( pSessionDetails->dwMaxPublicSlots != dwMaxPublicSlots  ||
                          pSessionDetails->dwMaxPrivateSlots != dwMaxPrivateSlots );

    // XSessionGetDetails only returns dwAvailablePublicSlots and dwAvailablePrivateSlots data
    // for:
    // 1. matchmaking sessions,
    // 2. sessions that are joinable during gameplay,
    // 3. sessions that aren't deleted
    //
    // Therefore it only makes sense to compare against are expected slot counts if all of
    // these conditions are met
    const BOOL bIsMatchmakingSession     = HasSessionFlags( XSESSION_CREATE_USES_MATCHMAKING );

    const BOOL bJoinable                 = !HasSessionFlags( XSESSION_CREATE_USES_ARBITRATION ) &&
                                           !HasSessionFlags( XSESSION_CREATE_JOIN_IN_PROGRESS_DISABLED );

    if( bIsMatchmakingSession && 
        ( m_SessionState != SessionStateDeleted ) &&
        ( ( m_SessionState != SessionStateInGame ) || ( ( m_SessionState == SessionStateInGame ) && bJoinable ) ) )
    {
        bSlotCountsDiffer |= ( pSessionDetails->dwAvailablePublicSlots != dwAvailablePublicSlots ||
                               pSessionDetails->dwAvailablePrivateSlots != dwAvailablePrivateSlots );
    }

    // Any differences detected?
    if( !bSlotCountsDiffer )
    {
        goto Terminate;
    }

    // Spew expected/actual slot counts
    DebugSpew( "***************** DebugValidateExpectedSlotCounts *****************\n" \
                    "instance: %p\n" \
                    "sessionID: %016I64X\n" \
                    "m_hSession: %p\n" \
                    "bIsHost: %d\n" \
                    "sessionstate: %s\n" \
                    "dwMaxPublicSlots        expected: %d;    actual: %d\n" \
                    "dwMaxPrivateSlots       expected: %d;    actual: %d\n",
                    this,
                    XNKIDToInt64( sessionID ),
                    m_hSession,
                    m_bIsHost,
                    g_astrSessionState[m_SessionState],
                    dwMaxPublicSlots, pSessionDetails->dwMaxPublicSlots,
                    dwMaxPrivateSlots, pSessionDetails->dwMaxPrivateSlots );

    if( HasSessionFlags( XSESSION_CREATE_USES_MATCHMAKING ) )
    {
        DebugSpew( "dwAvailablePublicSlots  expected: %d;    actual: %d\n" \
                   "dwAvailablePrivateSlots expected: %d;    actual: %d\n",
                   dwAvailablePublicSlots, pSessionDetails->dwAvailablePublicSlots,
                   dwAvailablePrivateSlots, pSessionDetails->dwAvailablePrivateSlots );
    }

    if( bBreakOnDifferent & bSlotCountsDiffer )
    {
        DebugBreak();
    }

Terminate:
    if ( pSessionDetails )
    {
        delete [] pSessionDetails;
    }

    #endif

    return !bSlotCountsDiffer;
}

//------------------------------------------------------------------------
// Name: GetSessionCreationReason()
// Desc: Retrieves session creation reason
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
SessionCreationReason SessionManager<TaskQueueType, MessageRouterType>::GetSessionCreationReason( VOID ) const
{
    return m_SessionCreationReason;        
}

//------------------------------------------------------------------------
// Name: GetSessionState()
// Desc: Retrieves current session state
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
SessionState SessionManager<TaskQueueType, MessageRouterType>::GetSessionState( VOID ) const
{
    return m_SessionState;        
}

//------------------------------------------------------------------------
// Name: GetSessionStateString()
// Desc: Retrieves current session state
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
const char* SessionManager<TaskQueueType, MessageRouterType>::GetSessionStateString( VOID ) const
{
    return g_astrSessionState[m_SessionState];
}

//------------------------------------------------------------------------
// Name: SetSessionOwner()
// Desc: Sets current session owner
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
VOID SessionManager<TaskQueueType, MessageRouterType>::SetSessionOwner( DWORD dwOwnerController )
{
    m_dwOwnerController = dwOwnerController;
    XUserGetXUID( m_dwOwnerController, &m_xuidOwner );
}

//------------------------------------------------------------------------
// Name: GetSessionOwner()
// Desc: Retrieves current session owner
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
DWORD SessionManager<TaskQueueType, MessageRouterType>::GetSessionOwner( VOID ) const
{
    return m_dwOwnerController;
}

//------------------------------------------------------------------------
// Name: GetSessionOwner()
// Desc: Retrieves current session owner XUID
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
XUID SessionManager<TaskQueueType, MessageRouterType>::GetSessionOwnerXuid( VOID ) const
{
    return m_xuidOwner;
}

//------------------------------------------------------------------------
// Name: IsSessionOwner()
// Desc: Retrieves current session owner
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
BOOL SessionManager<TaskQueueType, MessageRouterType>::IsSessionOwner( DWORD dwController ) const
{
    return ( dwController == m_dwOwnerController );
}

//------------------------------------------------------------------------
// Name: IsSessionOwner()
// Desc: Retrieves current session owner
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
BOOL SessionManager<TaskQueueType, MessageRouterType>::IsSessionOwner( XUID xuidOwner ) const
{
    return ( xuidOwner == m_xuidOwner );
}

//------------------------------------------------------------------------
// Name: IsSessionHost()
// Desc: True if this SessionManager instance is the session host,
//       else false
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
BOOL SessionManager<TaskQueueType, MessageRouterType>::IsSessionHost( VOID ) const
{
    return m_bIsHost;
}

//------------------------------------------------------------------------
// Name: MakeSessionHost()
// Desc: Make this SessionManager instance the session host. 
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
VOID SessionManager<TaskQueueType, MessageRouterType>::MakeSessionHost( VOID )
{
    m_bIsHost = TRUE;
}

//------------------------------------------------------------------------
// Name: GetSessionNonce()
// Desc: Function to retrieve current session nonce
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
ULONGLONG SessionManager<TaskQueueType, MessageRouterType>::GetSessionNonce( VOID ) const
{
    return m_qwSessionNonce;
}

//------------------------------------------------------------------------
// Name: SetSessionNonce()
// Desc: Function to set the session nonce
//       Only non-hosts should ever call this method
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
VOID SessionManager<TaskQueueType, MessageRouterType>::SetSessionNonce( ULONGLONG qwNonce )
{    
    if( m_bIsHost )
    {
        FatalError( "Only non-hosts should call this function!\n" );
    }

    m_qwSessionNonce = qwNonce;

    DebugDumpSessionDetails();
}

//------------------------------------------------------------------------
// Name: SetHostInAddr()
// Desc: Sets the IN_ADDR of the session host
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
VOID SessionManager<TaskQueueType, MessageRouterType>::SetHostInAddr( const IN_ADDR& inaddr )
{
    m_HostInAddr = inaddr;
}

//------------------------------------------------------------------------
// Name: GetHostInAddr()
// Desc: Returns the IN_ADDR of the session host
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
IN_ADDR SessionManager<TaskQueueType, MessageRouterType>::GetHostInAddr( VOID ) const
{
    return m_HostInAddr;
}


//------------------------------------------------------------------------
// Name: GetSessionID()
// Desc: Function to retrieve current session ID
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
XNKID SessionManager<TaskQueueType, MessageRouterType>::GetSessionID( VOID )  const
{
    return m_SessionInfo.sessionID;
}

//------------------------------------------------------------------------
// Name: GetSessionIDAsInt()
// Desc: Function to retrieve current session ID as an __int64
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
__int64 SessionManager<TaskQueueType, MessageRouterType>::GetSessionIDAsInt( VOID )  const
{
    return *(const __int64*)&m_SessionInfo.sessionID;
}

//------------------------------------------------------------------------
// Name: GetMigratedSessionID()
// Desc: Function to retrieve the session ID of the session while
//       it is being migrated. This method can only be called
//       when this session manager instance is actively migrating a session
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
XNKID SessionManager<TaskQueueType, MessageRouterType>::GetMigratedSessionID( VOID )  const
{
    if( !( m_SessionState >= SessionStateMigrateHost &&
           m_SessionState <= SessionStateMigratedHost ) )
    {
        FatalError( "GetMigratedSessionID called in state %s!\n", GetSessionStateString() );
    }

    return m_migratedSessionID;
}


//------------------------------------------------------------------------
// Name: SetSessionState()
// Desc: Sets current session state
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
VOID SessionManager<TaskQueueType, MessageRouterType>::SetSessionState( SessionState newState )
{
    m_SessionState = newState;
}

//------------------------------------------------------------------------
// Name: CheckSessionState()
// Desc: Checks if session in the expected state
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
BOOL SessionManager<TaskQueueType, MessageRouterType>::CheckSessionState( SessionState expectedState )  const
{
    SessionState actual = GetSessionState();

    if( actual != expectedState )
    {
        FatalError( "SessionManager instance in invalid state. Expected: %d; Got: %d\n", 
                    xpectedState, actual );

        return FALSE;
    }

    return TRUE;
}

//------------------------------------------------------------------------
// Name: GetRegistrationResults()
// Desc: Retrieves last results from registering this client for
//          arbitratrion
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
const PXSESSION_REGISTRATION_RESULTS SessionManager<TaskQueueType, MessageRouterType>::GetRegistrationResults()  const
{
    return (const PXSESSION_REGISTRATION_RESULTS)m_pRegistrationResults;    
}

//------------------------------------------------------------------------
// Name: SwitchToState()
// Desc: Changes to a new session state and performs initialization 
//       for the new state
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
VOID SessionManager<TaskQueueType, MessageRouterType>::SwitchToState( SessionState newState )
{
    // Ignore transitions to the same state
    if( m_SessionState == newState )
    {
        return;
    }

    // If we're in state SessionStateMigratedHost, force a switch to 
    // the state we were in prior to migration
    if( m_SessionState == SessionStateMigratedHost )
    {
        DebugSpew( "SessionManager %016I64X (instance %p): Detected to be in state SessionStateMigratedHost. " \
                   "Switching to correct pre-migration state\n",
                   XNKIDToInt64( m_SessionInfo.sessionID ),
                   this );
        
        SwitchToPreHostMigrationState();
    }

    assert( m_SessionState < _countof( g_astrSessionState ) );
    DebugSpew( "SessionManager %016I64X (instance %p): Switching from session state %s to %s\n", 
               XNKIDToInt64( m_SessionInfo.sessionID ), 
               this, 
               g_astrSessionState[m_SessionState], 
               g_astrSessionState[newState] );

    // Clean up from the previous state
    switch( m_SessionState )
    {
        case SessionStateDeleted:
            //Reset();

            // Clear out our slots and update our presence info
            m_Slots[ SLOTS_FILLEDPUBLIC  ] = 0;
            m_Slots[ SLOTS_FILLEDPRIVATE ] = 0;

            if( m_pRegistrationResults )
            {
                delete[] m_pRegistrationResults;
                m_pRegistrationResults = NULL;
            }
            break;

        case SessionStateMigratedHost:
            m_bIsMigratedSessionHost = FALSE;
            break;

        case SessionStateRegistering:
        case SessionStateInGame:
        case SessionStateNone:
        case SessionStateCreating:
        case SessionStateWaitingForRegistration:
        case SessionStateStarting:
        case SessionStateEnding:
        case SessionStateDeleting:
        default:
            break;
    }

    // Initialize the next state
    switch( newState )
    {
        case SessionStateCreate:
            CreateSession();
            break;

        case SessionStateCreated:
            if( !IsSessionHost() )
            {
                // Re-entering this state is a bad thing, since
                // that would wipe out the session nonce
                assert( m_SessionState != SessionStateCreated );
                
                // XSessionCreate would have
                // returned a 0 session nonce,
                // so set our session nonce
                // to a random value that we'll
                // replace once we get the
                // true nonce from the session host
                int retVal = XNetRandom( (BYTE *)&m_qwSessionNonce, sizeof( m_qwSessionNonce ) );
                if ( retVal )
                {
                    FatalError( "XNetRandom failed: %d\n", retVal );
                }
                DebugDumpSessionDetails();
            }
            break;

        case SessionStateRegister:
            RegisterArbitration();
            break;

        case SessionStateStart:
            StartSession();
            break;

        case SessionStateEnd:
            EndSession();
            break;

        case SessionStateMigrateHost:
            // Cache session state before we begin migration. We'll
            // return this SessionManager instance to this state
            // after migration
            m_migratedSessionState = m_SessionState;

            // Begin migration
            MigrateSession();
            break;

        case SessionStateMigratedHost:
            // Copy m_NewSessionInfo into m_SessionInfo
            // and zero out m_NewSessionInfo
            m_SessionInfo = m_NewSessionInfo; 
            ZeroMemory( &m_NewSessionInfo, sizeof( XSESSION_INFO ) );
            break;

        case SessionStateDelete:
            StopQoSListener();
            DeleteSession();
            break;

        case SessionStateDeleted:
            if( ( m_hSession != INVALID_HANDLE_VALUE ) && 
                ( m_hSession != NULL ) )
            {
                XCloseHandle( m_hSession );
                m_hSession = NULL;
            }
            break;

        case SessionStateNone:
        case SessionStateIdle:
        case SessionStateWaitingForRegistration:
        case SessionStateRegistered:
        case SessionStateInGame:
        default:
            break;
    }

    m_SessionState = newState;
}

//------------------------------------------------------------------------
// Name: SwitchToPreHostMigrationState()
// Desc: Switches the current session state back to what it was
//       prior to host migration
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
VOID SessionManager<TaskQueueType, MessageRouterType>::SwitchToPreHostMigrationState()
{
    // We can only transition to this state from SessionStateMigratedHost
    assert( m_SessionState == SessionStateMigratedHost );

    // Zero out m_migratedSessionID
    ZeroMemory( &m_migratedSessionID, sizeof( XNKID ) );

    // Return session state to what it was before migration started
    // and set m_migratedSessionState to a blank state
    m_SessionState           = m_migratedSessionState;
    m_migratedSessionState   = SessionStateNone;
    m_bIsMigratedSessionHost = FALSE;

    DebugDumpSessionDetails();
}


//------------------------------------------------------------------------
// Name: HasSessionFlags()
// Desc: Checks if passed-in flags are set for the current session
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
BOOL SessionManager<TaskQueueType, MessageRouterType>::HasSessionFlags( DWORD dwFlags )  const
{
    BOOL bHasFlags = FALSE;
    
    // What flags in m_dwSessionFlags and dwFlags are different?
    DWORD dwDiffFlags = m_dwSessionFlags ^ dwFlags;

    // If none of dwDiffFlags are in dwFlags,
    // we have a match
    if( ( dwDiffFlags & dwFlags ) == 0 )
    {
        bHasFlags = TRUE;
    }

    return bHasFlags;
}

//------------------------------------------------------------------------
// Name: FlipSessionFlags()
// Desc: XORs the state of the current session's
//       flags with that of the passed-in flags
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
VOID SessionManager<TaskQueueType, MessageRouterType>::FlipSessionFlags( DWORD dwFlags )
{ 
    m_dwSessionFlags ^= dwFlags;

    DebugSpew( "FlipSessionFlags: New session flags: %d\n", m_dwSessionFlags );
}

//------------------------------------------------------------------------
// Name: ClearSessionFlags()
// Desc: Clear the passed-in flags for the
//       current session
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
VOID SessionManager<TaskQueueType, MessageRouterType>::ClearSessionFlags( DWORD dwFlags )
{ 
    m_dwSessionFlags &= ~dwFlags; 

    DebugSpew( "ClearSessionFlags: New session flags: %d\n", m_dwSessionFlags );
}

//------------------------------------------------------------------------
// Name: SetSessionInfo()
// Desc: Set session info data. This only be called outside of an
//       active session
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
VOID SessionManager<TaskQueueType, MessageRouterType>::SetSessionInfo( const XSESSION_INFO& session_info )
{
    SessionState actual = GetSessionState();
    if( m_SessionState > SessionStateCreating && m_SessionState < SessionStateDelete )
    {
        FatalError( "SessionManager instance in invalid state: %s\n", g_astrSessionState[actual] );
    }

    m_SessionInfo = session_info;
}

//------------------------------------------------------------------------
// Name: SetNewSessionInfo()
// Desc: Set session info data for the new session we will migrate this
//       session to. This only be called inside of an active session
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
VOID SessionManager<TaskQueueType, MessageRouterType>::SetNewSessionInfo( const XSESSION_INFO& session_info,
                                                                          const BOOL bIsNewSessionHost )
{
    SessionState actual = GetSessionState();
    if( m_SessionState < SessionStateCreated || m_SessionState >= SessionStateDelete )
    {
        FatalError( "SessionManager instance in invalid state: %d\n", actual );
    }

    m_NewSessionInfo         = session_info;
    m_bIsMigratedSessionHost = bIsNewSessionHost;

    // Cache current session ID for a possible host migration scenario..
    memcpy( &m_migratedSessionID, &m_SessionInfo.sessionID, sizeof( XNKID ) ); 
}

//------------------------------------------------------------------------
// Name: SetSessionFlags()
// Desc: Set session flags. Optionally first clears all currently set 
//       flags. This only be called outside of an active session
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
VOID SessionManager<TaskQueueType, MessageRouterType>::SetSessionFlags( DWORD dwFlags, BOOL fClearExisting  )
{
    SessionState actual = GetSessionState();
    if( m_SessionState > SessionStateCreating && m_SessionState < SessionStateDelete )
    {
        FatalError( "SessionManager instance in invalid state: %s\n", g_astrSessionState[actual] );
    }

    if( fClearExisting )
    {
        m_dwSessionFlags = 0;
    }

    m_dwSessionFlags |= dwFlags;

    DebugSpew( "SetSessionFlags: New session flags: %d\n", m_dwSessionFlags );
}

//------------------------------------------------------------------------
// Name: GetSessionFlags()
// Desc: Returns current session flags
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
DWORD SessionManager<TaskQueueType, MessageRouterType>::GetSessionFlags()  const
{
    return m_dwSessionFlags;
}

//------------------------------------------------------------------------
// Name: GetSessionInfo()
// Desc: Returns current XSESSION_INFO
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
const XSESSION_INFO& SessionManager<TaskQueueType, MessageRouterType>::GetSessionInfo() const
{
    return m_SessionInfo;
}

//------------------------------------------------------------------------
// Name: GetMaxSlotCounts()
// Desc: Returns current maximum slot counts
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
VOID SessionManager<TaskQueueType, MessageRouterType>::GetMaxSlotCounts( DWORD& dwMaxPublicSlots, DWORD& dwMaxPrivateSlots )  const
{
    dwMaxPublicSlots  = m_Slots[SLOTS_TOTALPUBLIC];
    dwMaxPrivateSlots = m_Slots[SLOTS_TOTALPRIVATE];

//    DebugValidateExpectedSlotCounts();
}

//------------------------------------------------------------------------
// Name: GetFilledSlotCounts()
// Desc: Returns current filled slot counts
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
VOID SessionManager<TaskQueueType, MessageRouterType>::GetFilledSlotCounts( DWORD& dwFilledPublicSlots, DWORD& dwFilledPrivateSlots ) const
{
    dwFilledPublicSlots  = m_Slots[SLOTS_FILLEDPUBLIC];
    dwFilledPrivateSlots = m_Slots[SLOTS_FILLEDPRIVATE];

//    DebugValidateExpectedSlotCounts();
}

//------------------------------------------------------------------------
// Name: SetMaxSlotCounts()
// Desc: Sets current maximum slot counts
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
VOID SessionManager<TaskQueueType, MessageRouterType>::
    SetMaxSlotCounts( DWORD dwMaxPublicSlots, DWORD dwMaxPrivateSlots )
{
    #ifdef _DEBUG
    DebugSpew( "SetMaxSlotCounts(%016I64X): Old: [%d, %d, %d, %d]\n",
               XNKIDToInt64( m_SessionInfo.sessionID ),
               m_Slots[SLOTS_TOTALPUBLIC],
               m_Slots[SLOTS_TOTALPRIVATE],
               m_Slots[SLOTS_FILLEDPUBLIC],
               m_Slots[SLOTS_FILLEDPRIVATE]);
    #endif

    m_Slots[SLOTS_TOTALPUBLIC]  = dwMaxPublicSlots;
    m_Slots[SLOTS_TOTALPRIVATE] = dwMaxPrivateSlots;

    #ifdef _DEBUG
    DebugSpew( "SetMaxSlotCounts(%016I64X): New: [%d, %d, %d, %d]\n",
               XNKIDToInt64( m_SessionInfo.sessionID ),
               m_Slots[SLOTS_TOTALPUBLIC],
               m_Slots[SLOTS_TOTALPRIVATE],
               m_Slots[SLOTS_FILLEDPUBLIC],
               m_Slots[SLOTS_FILLEDPRIVATE]);
    #endif
}

//------------------------------------------------------------------------
// Name: SetFilledSlotCounts()
// Desc: Sets current filled slot counts
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
VOID SessionManager<TaskQueueType, MessageRouterType>::
    SetFilledSlotCounts( DWORD dwFilledPublicSlots, DWORD dwFilledPrivateSlots )
{
    #ifdef _DEBUG
    DebugSpew( "SetFilledSlotCounts(%016I64X): Old: [%d, %d, %d, %d]\n",
               m_SessionInfo.sessionID,
               m_Slots[SLOTS_TOTALPUBLIC],
               m_Slots[SLOTS_TOTALPRIVATE],
               m_Slots[SLOTS_FILLEDPUBLIC],
               m_Slots[SLOTS_FILLEDPRIVATE]);               
    #endif

    m_Slots[SLOTS_FILLEDPUBLIC]  = dwFilledPublicSlots;
    m_Slots[SLOTS_FILLEDPRIVATE] = dwFilledPrivateSlots;

    #ifdef _DEBUG
    DebugSpew( "SetFilledSlotCounts(%016I64X): New: [%d, %d, %d, %d]\n",
               m_SessionInfo.sessionID,
               m_Slots[SLOTS_TOTALPUBLIC],
               m_Slots[SLOTS_TOTALPRIVATE],
               m_Slots[SLOTS_FILLEDPUBLIC],
               m_Slots[SLOTS_FILLEDPRIVATE]);
    #endif
}

//------------------------------------------------------------------------
// Name: GetSessionError()
// Desc: Returns session error string
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
WCHAR* SessionManager<TaskQueueType, MessageRouterType>::GetSessionError()  const
{
    return m_strSessionError;
}

//------------------------------------------------------------------------
// Name: SetSessionError()
// Desc: Sets session error string
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
VOID SessionManager<TaskQueueType, MessageRouterType>::SetSessionError( WCHAR* error )
{
    m_strSessionError = error;
}


//--------------------------------------------------------------------------------------
// Name: StartQoSListener()
// Desc: Turn on the Quality of Service Listener for the current session
//--------------------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
VOID SessionManager<TaskQueueType, MessageRouterType>::StartQoSListener( BYTE* data, UINT dataLen, DWORD bitsPerSec )
{
    DWORD flags = XNET_QOS_LISTEN_SET_DATA;
    if( bitsPerSec )
    {
        flags |= XNET_QOS_LISTEN_SET_BITSPERSEC;
    }
    if( !m_bUsingQoS )
    {
        flags |= XNET_QOS_LISTEN_ENABLE;
    }
    DWORD dwRet;
    dwRet = XNetQosListen( &( m_SessionInfo.sessionID ), data, dataLen, bitsPerSec, flags );
    if( ERROR_SUCCESS != dwRet )
    {
        FatalError( "Failed to start QoS listener, error 0x%08x\n", dwRet );
    }
    m_bUsingQoS = TRUE;
}


//--------------------------------------------------------------------------------------
// Name: StopQoSListener()
// Desc: Turn off the Quality of Service Listener for the current session
//--------------------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
VOID SessionManager<TaskQueueType, MessageRouterType>::StopQoSListener()
{
    if( m_bUsingQoS )
    {
        DWORD dwRet;
        dwRet = XNetQosListen( &( m_SessionInfo.sessionID ), NULL, 0, 0,
                               XNET_QOS_LISTEN_RELEASE );
        if( ERROR_SUCCESS != dwRet )
        {
            DebugSpew( "Warning: Failed to stop QoS listener, error 0x%08x\n", dwRet );
        }
        m_bUsingQoS = FALSE;
    }
}

//------------------------------------------------------------------------
// Name: CreateSession()
// Desc: Creates a session
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
HRESULT SessionManager<TaskQueueType, MessageRouterType>::CreateSession( )
{
    // Make sure our instance is in the right state
    SessionState actual = GetSessionState();
    if( actual != SessionStateNone && actual != SessionStateDeleted  )
    {
        FatalError( "Wrong state: %s\n", g_astrSessionState[actual] );
    }

    // Modify session flags if host
    if( m_bIsHost )
    {
        m_dwSessionFlags |= XSESSION_CREATE_HOST;
    }
    else
    {
        // Turn off host bit
        m_dwSessionFlags &= ~XSESSION_CREATE_HOST;
    }

    // If we're just in a Matchmaking session, we can only specify the 
    // XSESSION_CREATE_JOIN_IN_PROGRESS_DISABLED modifier
    if( !( m_dwSessionFlags & XSESSION_CREATE_USES_PRESENCE ) && 
         ( m_dwSessionFlags & XSESSION_CREATE_USES_MATCHMAKING ) && 
         ( m_dwSessionFlags & XSESSION_CREATE_MODIFIERS_MASK ) )
    {
        // turn off all modifiers
        m_dwSessionFlags &= ~XSESSION_CREATE_MODIFIERS_MASK;

        // turn on XSESSION_CREATE_JOIN_IN_PROGRESS_DISABLED
        m_dwSessionFlags |= XSESSION_CREATE_JOIN_IN_PROGRESS_DISABLED;
    }

    // Set up our message
    DWORD dwMessage                 = MSG_SESSION_CREATE;
    SessionCreateMsg* pMsg          = new SessionCreateMsg();
    pMsg->m_dwFlags                 = m_dwSessionFlags;
    pMsg->m_dwUserIndex             = m_dwOwnerController;
    pMsg->m_dwMaxPrivateSlots       = m_Slots[SLOTS_TOTALPRIVATE];
    pMsg->m_dwMaxPublicSlots        = m_Slots[SLOTS_TOTALPUBLIC];

    // Set up our task
    PushTask( dwMessage, pMsg );

    // Return ERROR_IO_PENDING since we're by design async 
    return HRESULT_FROM_WIN32( ERROR_IO_PENDING );
}

//------------------------------------------------------------------------
// Name: StartSession()
// Desc: Starts a session
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
HRESULT SessionManager<TaskQueueType, MessageRouterType>::StartSession()
{
    // Make sure our instance is in the right state
    SessionState actual = GetSessionState();
    if( actual != SessionStateRegistered && actual != SessionStateCreated &&
        actual != SessionStateIdle )
    {
        FatalError( "Wrong state: %s\n", g_astrSessionState[actual] );
    }

    // Set up our message
    DWORD dwMessage             = MSG_SESSION_START;
    SessionStartMsg* pMsg       = new SessionStartMsg();

    // Set up our task
    PushTask( dwMessage, pMsg );

    // Return ERROR_IO_PENDING since we're by design async 
    return HRESULT_FROM_WIN32( ERROR_IO_PENDING );
}

//------------------------------------------------------------------------
// Name: EndSession()
// Desc: Ends the current session
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
HRESULT SessionManager<TaskQueueType, MessageRouterType>::EndSession()
{
    // Make sure our instance is in the right state. If not,
    // do nothing
    SessionState actual = GetSessionState();
    if( actual != SessionStateInGame && actual != SessionStateStarting )
    {
        DebugSpew( "SessionManager instance in invalid state %s. Ignoring EndSession\n", 
                   g_astrSessionState[actual] );
        return S_OK;
    }

    // Set up our message
    DWORD dwMessage     = MSG_SESSION_END;
    SessionEndMsg* pMsg = new SessionEndMsg();

    // Set up our task
    PushTask( dwMessage, pMsg );

    // Return ERROR_IO_PENDING since we're by design async 
    return HRESULT_FROM_WIN32( ERROR_IO_PENDING );
}

//------------------------------------------------------------------------
// Name: MigrateSession()
// Desc: Migrates the current session
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
HRESULT SessionManager<TaskQueueType, MessageRouterType>::MigrateSession()
{
    // Make sure our instance is in the right state
    SessionState actual = GetSessionState();
    if( actual < SessionStateCreated || actual >= SessionStateDelete  )
    {
        FatalError( "Wrong state: %s\n", g_astrSessionState[actual] );
    }

    // Set up our message
    DWORD dwMessage         = MSG_SESSION_MIGRATE;
    SessionMigrateMsg* pMsg = new SessionMigrateMsg();

    // Set up our task
    PushTask( dwMessage, pMsg );

    // Return ERROR_IO_PENDING since we're by design async 
    return HRESULT_FROM_WIN32( ERROR_IO_PENDING );
}

//------------------------------------------------------------------------
// Name: ModifySessionFlags()
// Desc: Modifies session flags
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
DWORD SessionManager<TaskQueueType, MessageRouterType>::ModifySessionFlags( DWORD dwFlags, 
                                                                            BOOL bClearFlags )
{
    SessionState actual = GetSessionState();
    if( m_SessionState <= SessionStateCreating || m_SessionState >= SessionStateDelete )
    {
        FatalError( "SessionManager instance in invalid state: %s\n", g_astrSessionState[actual] );
    }

    // If no flags in the modifiers mask are specified, do nothing
    if( ( dwFlags & XSESSION_CREATE_MODIFIERS_MASK ) == 0 )
    {
        return ERROR_SUCCESS;
    }

    // Flags cannot be modified for arbitrated sessions
    if( m_dwSessionFlags & XSESSION_CREATE_USES_ARBITRATION )
    {
        return ERROR_SUCCESS;
    }

    // Add host bit to dwFlags if host; otherwise remove
    if( m_bIsHost )
    {
        dwFlags |= XSESSION_CREATE_HOST;
    }
    else
    {
        // Turn off host bit
        dwFlags &= ~XSESSION_CREATE_HOST;
    }

    // Set up our message
    DWORD dwMessage             = MSG_SESSION_MODIFYFLAGS;
    SessionModifyFlagsMsg* pMsg = new SessionModifyFlagsMsg();
    pMsg->m_dwFlags             = dwFlags;
    pMsg->m_bClearFlags         = bClearFlags;

    // Set up our task
    PushTask( dwMessage, pMsg );

    // Return ERROR_IO_PENDING since we're by design async 
    return ERROR_IO_PENDING;
}

//------------------------------------------------------------------------
// Name: ModifySkill()
// Desc: Modifies TrueSkill(TM)
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
HRESULT SessionManager<TaskQueueType, MessageRouterType>::ModifySkill( const DWORD dwNumXuids, const XUID* pXuids )
{
    SessionState actual = GetSessionState();
    if( m_SessionState <= SessionStateCreating || m_SessionState >= SessionStateDelete )
    {
        FatalError( "SessionManager instance in invalid state: %s\n", g_astrSessionState[actual] );
    }

    // Set up our message
    DWORD dwMessage = MSG_SESSION_MODIFYSKILL;
    ModifySkillMsg* pMsg = new ModifySkillMsg( dwNumXuids, pXuids );

    // Set up our task
    PushTask( dwMessage, pMsg );

    // Return ERROR_IO_PENDING since we're by design async 
    return HRESULT_FROM_WIN32( ERROR_IO_PENDING );
}

//------------------------------------------------------------------------
// Name: RegisterArbitration()
// Desc: Registers the current session for arbitration
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
HRESULT SessionManager<TaskQueueType, MessageRouterType>::RegisterArbitration()
{
    SessionState actual = GetSessionState();
    if( actual < SessionStateCreated || actual >= SessionStateStarting )
    {
        FatalError( "Wrong state: %s\n", g_astrSessionState[actual] );
    }

    // Set up our message
    DWORD dwMessage                 = MSG_SESSION_REGISTERARBITRATION;
    ArbitrationRegisterMsg* pMsg    = new ArbitrationRegisterMsg();

    // Set up our task
    PushTask( dwMessage, pMsg );

    // Return ERROR_IO_PENDING since we're by design async 
    return HRESULT_FROM_WIN32( ERROR_IO_PENDING );
}

//------------------------------------------------------------------------
// Name: DeleteSession()
// Desc: Deletes the current session
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
HRESULT SessionManager<TaskQueueType, MessageRouterType>::DeleteSession()
{
    // Make sure our instance is in the right state
    SessionState actual = GetSessionState();
    if( actual < SessionStateCreated || actual > SessionStateDelete  )
    {
        FatalError( "Wrong state: %s\n", g_astrSessionState[actual] );
    }

    // Set up our message
    DWORD dwMessage         = MSG_SESSION_DELETE;
    SessionDeleteMsg* pMsg  = new SessionDeleteMsg();

    // Set up our task
    PushTask( dwMessage, pMsg );

    // Return ERROR_IO_PENDING since we're by design async 
    return HRESULT_FROM_WIN32( ERROR_IO_PENDING );
}

//------------------------------------------------------------------------
// Name: AddLocalPlayers()
// Desc: Add local players to the session
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
HRESULT SessionManager<TaskQueueType, MessageRouterType>::AddLocalPlayers( const DWORD dwUserCount,
                                         const DWORD* pdwUserIndices,
                                         const BOOL* pfPrivateSlots )
{
    SessionState actual = GetSessionState();
    if( actual < SessionStateCreating || actual >= SessionStateDelete )
    {
        FatalError( "Wrong state: %s\n", g_astrSessionState[actual] );
    }

    // Set up our message
    DWORD dwMessage             = MSG_SESSION_ADDLOCAL;
    AddLocalPlayersMsg* pMsg    = new AddLocalPlayersMsg( dwUserCount, pdwUserIndices, pfPrivateSlots );

    // Set up our task
    PushTask( dwMessage, pMsg );

    // Return ERROR_IO_PENDING since we're by design async 
    return HRESULT_FROM_WIN32( ERROR_IO_PENDING );
}

//------------------------------------------------------------------------
// Name: AddRemotePlayers()
// Desc: Adds remote players to the session
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
HRESULT SessionManager<TaskQueueType, MessageRouterType>::AddRemotePlayers( const DWORD dwXuidCount,
                                          const XUID* pXuids,
                                          const BOOL* pfPrivateSlots )
{
    SessionState actual = GetSessionState();
    if( actual < SessionStateCreating || actual >= SessionStateDelete )
    {
        FatalError( "Wrong state: %s\n", g_astrSessionState[actual] );
    }

    for ( DWORD i=0; i<dwXuidCount; i++ )
    {
        DebugSpew( "AddRemotePlayers: %I64u; bInvited: %d\n", pXuids[ i ], pfPrivateSlots[ i ] );
    }

    // Set up our message
    DWORD dwMessage             = MSG_SESSION_ADDREMOTE;
    AddRemotePlayersMsg* pMsg   = new AddRemotePlayersMsg( dwXuidCount, pXuids, pfPrivateSlots );

    // Set up our task
    PushTask( dwMessage, pMsg );

    // Return ERROR_IO_PENDING since we're by design async 
    return HRESULT_FROM_WIN32( ERROR_IO_PENDING );
}

//------------------------------------------------------------------------
// Name: RemoveLocalPlayers()
// Desc: Add local players to the session
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
HRESULT SessionManager<TaskQueueType, MessageRouterType>::RemoveLocalPlayers( const DWORD dwUserCount,
                                            const DWORD* pdwUserIndices,
                                            const BOOL* pfPrivateSlots )
{
    SessionState actual = GetSessionState();
    if( actual < SessionStateCreated || actual >= SessionStateDelete )
    {
        FatalError( "Wrong state: %s\n", g_astrSessionState[actual] );
    }

    // Set up our message
    DWORD dwMessage                 = MSG_SESSION_REMOVELOCAL;
    RemoveLocalPlayersMsg* pMsg     = new RemoveLocalPlayersMsg( dwUserCount, pdwUserIndices, pfPrivateSlots );

    // Set up our task
    PushTask( dwMessage, pMsg );

    // Return ERROR_IO_PENDING since we're by design async 
    return HRESULT_FROM_WIN32( ERROR_IO_PENDING );
}

//------------------------------------------------------------------------
// Name: RemoveRemotePlayers()
// Desc: Remove remote players to the session
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
HRESULT SessionManager<TaskQueueType, MessageRouterType>::RemoveRemotePlayers( const DWORD dwXuidCount,
                                             const XUID* pXuids,
                                             const BOOL* pfPrivateSlots )
{
    SessionState actual = GetSessionState();
    if( actual < SessionStateCreated || actual >= SessionStateDelete )
    {
        FatalError( "Wrong state: %s\n", g_astrSessionState[actual] );
    }

    // Set up our message
    DWORD dwMessage                 = MSG_SESSION_REMOVEREMOTE;
    RemoveRemotePlayersMsg* pMsg    = new RemoveRemotePlayersMsg( dwXuidCount, pXuids, pfPrivateSlots );

    // Set up our task
    PushTask( dwMessage, pMsg );

    // Return ERROR_IO_PENDING since we're by design async 
    return HRESULT_FROM_WIN32( ERROR_IO_PENDING );
}


//------------------------------------------------------------------------
// Name: WriteStats()
// Desc: Write stats for a player in the session
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
HRESULT SessionManager<TaskQueueType, MessageRouterType>::WriteStats( XUID xuid,
                                       DWORD dwNumViews,
                                       const XSESSION_VIEW_PROPERTIES *pViews )
{
    SessionState actual = GetSessionState();
    if( actual < SessionStateInGame || actual >= SessionStateEnding )
    {
        FatalError( "Wrong state: %s\n", g_astrSessionState[actual] );
    }

    // Set up our message
    DWORD dwMessage                 = MSG_SESSION_WRITESTATS;
    WriteStatsMsg* pMsg             = new WriteStatsMsg(xuid, dwNumViews, pViews );

    // Set up our task to execute immediately
    RunTaskNow( dwMessage, pMsg );

    // Return ERROR_IO_PENDING since we're by design async 
    return HRESULT_FROM_WIN32( ERROR_IO_PENDING );
}

//------------------------------------------------------------------------
// Name: ProcessCreateSessionMessage()
// Desc: Creates a session using the XSessionCreate API
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
DWORD SessionManager<TaskQueueType, MessageRouterType>::ProcessCreateSessionMessage( MessageBase* pMessage )
{
    // Create local copy of our message data
    SessionCreateMsg msg = *( SessionCreateMsg* )pMessage;
    
    // ... and free message memory
    pMessage->Release();

    InitializeOverlapped();
    
    DWORD ret = XSessionCreate( msg.m_dwFlags,
                                msg.m_dwUserIndex,
                                msg.m_dwMaxPublicSlots,
                                msg.m_dwMaxPrivateSlots,
                                &m_qwSessionNonce,
                                &m_SessionInfo,
                                &m_XOverlapped,
                                &m_hSession );

    if( ret != ERROR_IO_PENDING )
    {
        FatalError( "XSessionCreate failed with error %d\n", ret );
    }

    SwitchToState( SessionStateCreating );

    return ret;
}

//------------------------------------------------------------------------
// Name: ProcessStartSessionMessage()
// Desc: Creates a session using the XSessionStart API
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
DWORD SessionManager<TaskQueueType, MessageRouterType>::ProcessStartSessionMessage( MessageBase* pMessage )
{
    // free message memory
    pMessage->Release();

    InitializeOverlapped();

    DWORD ret = XSessionStart( m_hSession,
                               0,
                               &m_XOverlapped );

    if( ret != ERROR_IO_PENDING )
    {
        FatalError( "XSessionStart failed with error %d\n", ret );
    }

    SwitchToState( SessionStateStarting );
    
    return ret;
}

//------------------------------------------------------------------------
// Name: ProcessMigrateHostMessage()
// Desc: Migrates session to new host using the XSessionMigrateHost API
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
DWORD SessionManager<TaskQueueType, MessageRouterType>::ProcessMigrateHostMessage( MessageBase* pMessage )
{
    // free message memory
    pMessage->Release();

    InitializeOverlapped();

    DWORD ret = XSessionMigrateHost( m_hSession,
                                     ( m_bIsMigratedSessionHost ) ? m_dwOwnerController : XUSER_INDEX_NONE,
                                     &m_NewSessionInfo,
                                     &m_XOverlapped );

    if( ret != ERROR_IO_PENDING )
    {
        FatalError( "XSessionMigrateHost failed with error %d\n", ret );
    }

    SwitchToState( SessionStateMigratingHost );

    return ret;
}

//------------------------------------------------------------------------
// Name: ProcessWriteStatsMessage()
// Desc: Writes stats for the current session
//       using the XSessionWriteStats API
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
DWORD SessionManager<TaskQueueType, MessageRouterType>::ProcessWriteStatsMessage( MessageBase* pMessage )
{
    DWORD ret = ERROR_SUCCESS;

    // Create local copy of our message data
    WriteStatsMsg msg = *( WriteStatsMsg* )pMessage;
    
    // ... and free message memory
    pMessage->Release();

    SessionState actual = GetSessionState();
    if( actual > SessionStateEnding )
    {
        // Can only write stats between XSessionStart and XSessionEnd, so
        // do nothing
        return ret;
    }

    DebugSpew( "%016I64X: ProcessWriteStatsMessage: %I64u\n", 
               GetSessionIDAsInt(), 
               msg.m_xuid );

    InitializeOverlapped();

    ret = XSessionWriteStats( m_hSession,
                              msg.m_xuid,
                              msg.m_dwNumViews,
                              msg.m_pViews,
                              &m_XOverlapped );

    if( ret != ERROR_IO_PENDING )
    {
        FatalError( "XSessionWriteStats failed with error %d\n", ret );
    }

    return ret;
}


//------------------------------------------------------------------------
// Name: ProcessEndSessionMessage()
// Desc: Ends current session using the XSessionEnd API
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
DWORD SessionManager<TaskQueueType, MessageRouterType>::ProcessEndSessionMessage( MessageBase* pMessage )
{
    // free message memory
    pMessage->Release();
    
    InitializeOverlapped();

    DWORD ret = XSessionEnd( m_hSession, &m_XOverlapped );

    if( ret != ERROR_IO_PENDING )
    {
        FatalError( "XSessionEnd failed with error %d\n", ret );
    }

    SwitchToState( SessionStateEnding );

    return ret;
}

//------------------------------------------------------------------------
// Name: ProcessDeleteSessionMessage()
// Desc: Deletes current session using the XSessionDelete API
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
DWORD SessionManager<TaskQueueType, MessageRouterType>::ProcessDeleteSessionMessage( MessageBase* pMessage )
{
    // free message memory
    pMessage->Release();

    SessionState actual = GetSessionState();
    if( actual > SessionStateDelete || IsSessionDeleted() )
    {
        FatalError( "Session already in a deleting state: %s\n", 
                    g_astrSessionState[actual] );
    }

    InitializeOverlapped();

    DWORD ret = XSessionDelete( m_hSession, &m_XOverlapped );

    if( ret != ERROR_IO_PENDING )
    {
        FatalError( "XSessionDelete failed with error %d\n", ret );
    }

    SwitchToState( SessionStateDeleting );

    return ret;
}

//------------------------------------------------------------------------
// Name: ProcessModifySessionFlagsMessage()
// Desc: Modifies a session's flags using the XSessionModify API
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
DWORD SessionManager<TaskQueueType, MessageRouterType>::ProcessModifySessionFlagsMessage( MessageBase* pMessage )
{
    // Create local copy of our message data
    SessionModifyFlagsMsg msg = *( SessionModifyFlagsMsg* )pMessage;
    
    // ... and free message memory
    pMessage->Release();

    // Flags cannot be modified for arbitrated sessions
    if( m_dwSessionFlags & XSESSION_CREATE_USES_ARBITRATION )
    {
        return ERROR_SUCCESS;
    }

    InitializeOverlapped();

    // Apply allowable session modifier flags to our session flags
    if( msg.m_bClearFlags )
    {
        m_dwSessionFlags ^= msg.m_dwFlags & MODIFY_FLAGS_ALLOWED;
    }
    else
    {
        m_dwSessionFlags |= msg.m_dwFlags & MODIFY_FLAGS_ALLOWED;
    }

    // Re-insert host bit if host
    if( m_bIsHost )
    {
        m_dwSessionFlags |= XSESSION_CREATE_HOST;
    }

    DebugSpew( "ProcessModifySessionFlagsMessage: New session flags: %d\n", m_dwSessionFlags );

    DWORD ret = XSessionModify( m_hSession,
                                m_dwSessionFlags,
                                m_Slots[SLOTS_TOTALPUBLIC],
                                m_Slots[SLOTS_TOTALPRIVATE],
                                &m_XOverlapped );

    if( ret != ERROR_IO_PENDING )
    {
        FatalError( "XSessionModify failed with error %d\n", ret );
    }

    // Wait for the overlapped operation to complete
    while ( IsProcessingOverlappedOperation() );

    HRESULT hr = HRESULT_FROM_WIN32( XGetOverlappedExtendedError( &m_XOverlapped ) );
    if( !SUCCEEDED( hr ) )
    {
        FatalError( "XGetOverlappedExtendedError returned error 0x%08x\n", hr );
    }

    DebugDumpSessionDetails();

    return ret;
}

//------------------------------------------------------------------------
// Name: ProcessArbitrationRegisterMessage()
// Desc: Deletes current session using the XSessionArbitrationRegister API
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
DWORD SessionManager<TaskQueueType, MessageRouterType>::ProcessArbitrationRegisterMessage( MessageBase* pMessage )
{
    // free message memory
    pMessage->Release();

    // Call once to determine size of results buffer
    DWORD cbRegistrationResults = 0;

    DWORD ret = XSessionArbitrationRegister( m_hSession,
                                             0,
                                             m_qwSessionNonce,
                                             &cbRegistrationResults,
                                             NULL,
                                             NULL ); 

    if( ( ret != ERROR_INSUFFICIENT_BUFFER ) || ( cbRegistrationResults == 0 ) )
    {
        FatalError( "Failed on first call to XSessionArbitrationRegister, hr=0x%08x\n", ret );
    }

    m_pRegistrationResults = (PXSESSION_REGISTRATION_RESULTS)new BYTE[cbRegistrationResults];
    if( m_pRegistrationResults == NULL )
    {
        FatalError( "Failed to allocate buffer.\n" );
    }

    InitializeOverlapped();

    // Call second time to fill our results buffer
    ret = XSessionArbitrationRegister( m_hSession,
                                       0,
                                       m_qwSessionNonce,
                                       &cbRegistrationResults,
                                       m_pRegistrationResults,
                                       &m_XOverlapped ); 

    if( ret != ERROR_IO_PENDING )
    {
        FatalError( "XSessionArbitrationRegister failed with error %d\n", ret );
    }

    SwitchToState( SessionStateRegistering );

    return ret;
}

//------------------------------------------------------------------------
// Name: ProcessAddLocalPlayerMessage()
// Desc: Add local players using the XSessionJoinLocal API
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
DWORD SessionManager<TaskQueueType, MessageRouterType>::ProcessAddLocalPlayerMessage( MessageBase* pMessage )
{
    DWORD ret = ERROR_SUCCESS;

    // Create local copy of our message data
    AddLocalPlayersMsg msg = *( AddLocalPlayersMsg* )pMessage;

    // ... and free message memory
    pMessage->Release();

    // Do nothing if the session is already deleted
    if( m_SessionState == SessionStateDeleted )
    {
        return ret;
    }

    for ( DWORD  i = 0; i < msg.m_dwUserCount; i++ )
    {
        XUID xuid;
        XUserGetXUID( msg.m_pdwUserIndices[i], &xuid );

        DebugSpew( "Processing local user: index: %d; xuid: %I64u; bInvited: %d\n", 
                   msg.m_pdwUserIndices[ i ], xuid, msg.m_pfPrivateSlots[ i ] );

        // If xuid is already in the session, don't re-add
        if( IsPlayerInSession( xuid ) )
        {
            msg.m_pdwUserIndices[ i ] = XUSER_INDEX_NONE;
        }
    }

    InitializeOverlapped();

    ret = XSessionJoinLocal( m_hSession,
                             msg.m_dwUserCount,
                             msg.m_pdwUserIndices,
                             msg.m_pfPrivateSlots,
                             &m_XOverlapped );

    if( ret != ERROR_IO_PENDING )
    {
        FatalError( "XSessionJoinLocal failed with error %d\n", ret );
    }

    // Wait for the overlapped operation to complete
    while ( IsProcessingOverlappedOperation() );

    // Make sure we were able to add local users to the
    // session and then update slot counts
    ret = XGetOverlappedExtendedError( &m_XOverlapped );
    if( ret != ERROR_SUCCESS )
    {
        DebugSpew( "Failed to add remote users to session, error 0x%08x\n", HRESULT_FROM_WIN32( ret ) );
        return ret;
    }
    else
    {
        for ( DWORD i = 0; i< msg.m_dwUserCount; i++ )
        {
            m_Slots[ SLOTS_FILLEDPRIVATE ] += ( msg.m_pfPrivateSlots[ i ] ) ? 1 : 0;
            m_Slots[ SLOTS_FILLEDPUBLIC ] += ( msg.m_pfPrivateSlots[ i ] ) ? 0 : 1;
        }

        if( m_Slots[ SLOTS_FILLEDPRIVATE ] > m_Slots[ SLOTS_TOTALPRIVATE ] )
        {
            DWORD diff = m_Slots[ SLOTS_FILLEDPRIVATE ] - m_Slots[ SLOTS_TOTALPRIVATE ];
            m_Slots[ SLOTS_FILLEDPUBLIC ] += diff;

            assert( m_Slots[ SLOTS_FILLEDPUBLIC ] <= m_Slots[ SLOTS_TOTALPUBLIC ] );
            if( m_Slots[ SLOTS_FILLEDPUBLIC ] > m_Slots[ SLOTS_TOTALPUBLIC ] )
            {
                FatalError( "Too many slots filled!\n" );
            }

            m_Slots[ SLOTS_FILLEDPRIVATE ] = m_Slots[ SLOTS_TOTALPRIVATE ];
        }
    }

    DebugDumpSessionDetails();
    DebugValidateExpectedSlotCounts();

    return ret;
}

//------------------------------------------------------------------------
// Name: ProcessAddRemotePlayerMessage()
// Desc: Add local players using the XSessionJoinRemote API
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
DWORD SessionManager<TaskQueueType, MessageRouterType>::ProcessAddRemotePlayerMessage( MessageBase* pMessage )
{
    DWORD ret = ERROR_SUCCESS;

    // Create local copy of our message data
    AddRemotePlayersMsg msg = *( AddRemotePlayersMsg* )pMessage;

    // ... and free message memory
    pMessage->Release();

    // Do nothing if the session is already deleted
    if( m_SessionState == SessionStateDeleted )
    {
        return ret;
    }

    for ( DWORD  i = 0; i < msg.m_dwXuidCount; i++ )
    {
        DebugSpew( "Processing remote user: %I64u; bInvited: %d\n", 
                   msg.m_pXuids[ i ], msg.m_pfPrivateSlots[ i ] );

        // If msg.m_pXuids[ i ] is already in the session, don't re-add
        if( IsPlayerInSession( msg.m_pXuids[ i ] ) )
        {
            msg.m_pXuids[ i ] = INVALID_XUID;
        }
    }

    InitializeOverlapped();

    ret = XSessionJoinRemote( m_hSession,
                              msg.m_dwXuidCount,
                              msg.m_pXuids,
                              msg.m_pfPrivateSlots,
                              &m_XOverlapped );

    if( ret != ERROR_IO_PENDING )
    {
        FatalError( "XSessionJoinRemote failed with error %d\n", ret );
    }

    // Wait for the overlapped operation to complete
    while ( IsProcessingOverlappedOperation() );

    // Make sure we were able to remote local users to the
    // session and then update slot counts
    ret = XGetOverlappedExtendedError( &m_XOverlapped );
    if( ret != ERROR_SUCCESS )
    {
        DebugSpew( "Failed to add remote users to session, error 0x%08x\n", HRESULT_FROM_WIN32( ret ) );
        return ret;
    }
    else
    {
        for ( DWORD i = 0; i < msg.m_dwXuidCount; i++ )
        {
            m_Slots[ SLOTS_FILLEDPRIVATE ] += ( msg.m_pfPrivateSlots[ i ] ) ? 1 : 0;
            m_Slots[ SLOTS_FILLEDPUBLIC ] += ( msg.m_pfPrivateSlots[ i ] ) ? 0 : 1;
        }

        if( m_Slots[ SLOTS_FILLEDPRIVATE ] > m_Slots[ SLOTS_TOTALPRIVATE ] )
        {
            DWORD diff = m_Slots[ SLOTS_FILLEDPRIVATE ] - m_Slots[ SLOTS_TOTALPRIVATE ];
            m_Slots[ SLOTS_FILLEDPUBLIC ] += diff;

            assert( m_Slots[ SLOTS_FILLEDPUBLIC ] <= m_Slots[ SLOTS_TOTALPUBLIC ] );
            if( m_Slots[ SLOTS_FILLEDPUBLIC ] > m_Slots[ SLOTS_TOTALPUBLIC ] )
            {
                FatalError( "Too many slots filled!\n" );
            }

            m_Slots[ SLOTS_FILLEDPRIVATE ] = m_Slots[ SLOTS_TOTALPRIVATE ];
        }
    }

    DebugDumpSessionDetails();
    DebugValidateExpectedSlotCounts();

    return ret;
}

//------------------------------------------------------------------------
// Name: ProcessRemoveLocalPlayerMessage()
// Desc: Remove local players using the XSessionLeaveLocal API
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
DWORD SessionManager<TaskQueueType, MessageRouterType>::ProcessRemoveLocalPlayerMessage( MessageBase* pMessage )
{
    DWORD ret = ERROR_SUCCESS;

    // Create local copy of our message data
    RemoveLocalPlayersMsg msg = *( RemoveLocalPlayersMsg* )pMessage;

    // ... and free message memory
    pMessage->Release();

    // Do nothing if the session is already deleted
    if( m_SessionState == SessionStateDeleted )
    {
        return ret;
    }

    InitializeOverlapped();

    #ifdef _DEBUG
    for( DWORD i = 0; i < msg.m_dwUserCount; i++ )
    {
        XUID xuid;
        XUserGetXUID( i, &xuid );
        DebugSpew( "%016I64X: ProcessRemoveLocalPlayerMessage: %I64u\n", 
                   GetSessionIDAsInt(), 
                   xuid );
    }
    #endif

    ret = XSessionLeaveLocal( m_hSession,
                              msg.m_dwUserCount,
                              msg.m_pdwUserIndices,
                              &m_XOverlapped );

    if( ret != ERROR_IO_PENDING )
    {
        FatalError( "XSessionLeaveLocal failed with error %d\n", ret );
    }

    // Wait for the overlapped operation to complete
    while ( IsProcessingOverlappedOperation() );

    // Make sure we were able to remove local users from the
    // session and then update slot counts
    ret = XGetOverlappedExtendedError( &m_XOverlapped );
    if( ret != ERROR_SUCCESS )
    {
        FatalError( "Failed to remove local users from session, error 0x%08x\n", HRESULT_FROM_WIN32( ret ) );
    }
    else
    {
        // If the session is arbitrated and gameplay is happening, then the player will still be
        // kept in the session so stats can be reported, so we don't want to modify our slot counts
        const BOOL bPlayersKeptForStatsReporting = ( ( m_dwSessionFlags & XSESSION_CREATE_USES_ARBITRATION ) && 
                                                     ( m_SessionState == SessionStateInGame ) );

        if( !bPlayersKeptForStatsReporting )
        {
            for ( DWORD i = 0; i < msg.m_dwUserCount; i++ )
            {
                m_Slots[ SLOTS_FILLEDPRIVATE ] -= ( msg.m_pfPrivateSlots[ i ] ) ? 1 : 0;
                m_Slots[ SLOTS_FILLEDPUBLIC ] -= ( msg.m_pfPrivateSlots[ i ] ) ? 0 : 1;

                m_Slots[ SLOTS_FILLEDPRIVATE ] = max( m_Slots[ SLOTS_FILLEDPRIVATE ], 0 );
                m_Slots[ SLOTS_FILLEDPUBLIC ] = max( m_Slots[ SLOTS_FILLEDPUBLIC ], 0 );
            }
        }
        else
        {
            DebugSpew( "%016I64X: ProcessRemoveLocalPlayerMessage: Arbitrated session, so leaving players become zombies and not updating local slot counts\n", 
                       GetSessionIDAsInt() );
        }
    }

    DebugDumpSessionDetails();
    DebugValidateExpectedSlotCounts();

    return ret;
}

//------------------------------------------------------------------------
// Name: ProcessRemoveRemotePlayerMessage()
// Desc: Add local players using the XSessionLeaveRemote API
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
DWORD SessionManager<TaskQueueType, MessageRouterType>::ProcessRemoveRemotePlayerMessage( MessageBase* pMessage )
{
    DWORD ret = ERROR_SUCCESS;

    // Create local copy of our message data
    RemoveRemotePlayersMsg msg = *( RemoveRemotePlayersMsg* )pMessage;

    // ... and free message memory
    pMessage->Release();

    // Do nothing if the session is already deleted
    if( m_SessionState == SessionStateDeleted )
    {
        return ret;
    }

    InitializeOverlapped();

    #ifdef _DEBUG
    for( DWORD i = 0; i < msg.m_dwXuidCount; i++ )
    {
        DebugSpew( "%016I64X: ProcessRemoveRemotePlayerMessage: %I64u\n", 
                   GetSessionIDAsInt(), 
                   msg.m_pXuids[i] );
    }
    #endif

    ret = XSessionLeaveRemote( m_hSession,
                               msg.m_dwXuidCount,
                               msg.m_pXuids,
                               &m_XOverlapped );

    if( ret != ERROR_IO_PENDING )
    {
        FatalError( "XSessionLeaveRemote failed with error %d\n", ret );
    }

    DebugDumpSessionDetails();

    // Wait for the overlapped operation to complete
    while ( IsProcessingOverlappedOperation() );

    // Make sure we were able to remove local users from the
    // session and then update slot counts
    ret = XGetOverlappedExtendedError( &m_XOverlapped );
    if( ret != ERROR_SUCCESS )
    {
        FatalError( "Failed to add local users to session, error 0x%08x\n", HRESULT_FROM_WIN32( ret ) );
    }
    else
    {
        // If the session is arbitrated and gameplay is happening, then the player will still be
        // kept in the session so stats can be reported, so we don't want to modify our slot counts
        const BOOL bPlayersKeptForStatsReporting = ( ( m_dwSessionFlags & XSESSION_CREATE_USES_ARBITRATION ) && 
                                                     ( m_SessionState == SessionStateInGame ) );

        if( !bPlayersKeptForStatsReporting )
        {
            for ( DWORD i = 0; i < msg.m_dwXuidCount; i++ )
            {
                m_Slots[ SLOTS_FILLEDPRIVATE ] -= ( msg.m_pfPrivateSlots[ i ] ) ? 1 : 0;
                m_Slots[ SLOTS_FILLEDPUBLIC ] -= ( msg.m_pfPrivateSlots[ i ] ) ? 0 : 1;

                m_Slots[ SLOTS_FILLEDPRIVATE ] = max( m_Slots[ SLOTS_FILLEDPRIVATE ], 0 );
                m_Slots[ SLOTS_FILLEDPUBLIC ] = max( m_Slots[ SLOTS_FILLEDPUBLIC ], 0 );
            }
        }
        else
        {
            DebugSpew( "%016I64X: ProcessRemoveRemotePlayerMessage: Arbitrated session, so leaving players become zombies and not updating local slot counts\n", 
                       GetSessionIDAsInt() );
        }
    }

    DebugDumpSessionDetails();
    DebugValidateExpectedSlotCounts();

    return ret;
}

//------------------------------------------------------------------------
// Name: ProcessModifySkillMessage()
// Desc: Modifies the TrueSkill(TM) of a session using the 
//       XSessionModifySkill API
//------------------------------------------------------------------------
template <class TaskQueueType, class MessageRouterType>
DWORD SessionManager<TaskQueueType, MessageRouterType>::ProcessModifySkillMessage( MessageBase* pMessage )
{
    DWORD ret = ERROR_SUCCESS;

    // Create local copy of our message data
    ModifySkillMsg msg = *( ModifySkillMsg* )pMessage;

    // ... and free message memory
    pMessage->Release();

    // Do nothing if the session is already being deleted
    if( m_SessionState >= SessionStateDelete || IsSessionDeleted() )
    {
        return ret;
    }

    InitializeOverlapped();

    ret = XSessionModifySkill( m_hSession,
                               msg.m_dwXuidCount,
                               msg.m_pXuids,
                               &m_XOverlapped );

    if( ret != ERROR_IO_PENDING )
    {
        FatalError( "XSessionModifySkill failed with error %d\n", ret );
    }

    // Wait for the overlapped operation to complete
    while ( IsProcessingOverlappedOperation() );

    ret = XGetOverlappedExtendedError( &m_XOverlapped );
    if( ret != ERROR_SUCCESS )
    {
        FatalError( "XSessionModifySkill failed with error 0x%08x\n", HRESULT_FROM_WIN32( ret ) );
    }

    return ret;
}

// Force instantiation of SessionManagerMessageCallback with a specific template
static PFNMESSAGECONSUMERCALLBACK g_pfn = 
    SessionManagerMessageCallback<TaskQueue,MessageRouter<TaskQueue>>;

}
